#!/usr/bin/env python
from conan import ConanFile
from conan.tools.cmake import CMake, CMakeToolchain, CMakeDeps, cmake_layout
from conan.tools.files import copy, rmdir
from conan.tools.build import check_min_cppstd
from conan.tools.scm import Version
from conan.errors import ConanInvalidConfiguration
import os
import re

required_conan_version = ">=1.53.0"

class CatchConan(ConanFile):
    name = "catch2"
    description = "A modern, C++-native, framework for unit-tests, TDD and BDD"
    topics = ("conan", "catch2", "unit-test", "tdd", "bdd")
    url = "https://github.com/catchorg/Catch2"
    homepage = url
    license = "BSL-1.0"
    version = "latest"
    settings = "os", "compiler", "build_type", "arch"
    extension_properties = {"compatibility_cppstd": False}

    options = {
        "shared": [True, False],
        "fPIC": [True, False],
    }
    default_options = {
        "shared": False,
        "fPIC": True,
    }

    @property
    def _min_cppstd(self):
        return "14"

    @property
    def _compilers_minimum_version(self):
        return {
            "gcc": "7",
            "Visual Studio": "15",
            "msvc": "191",
            "clang": "5",
            "apple-clang": "10",
        }


    def set_version(self):
        pattern = re.compile(r"\w*VERSION (\d+\.\d+\.\d+) # CML version placeholder, don't delete")
        with open("CMakeLists.txt") as file:
            for line in file:
                result = pattern.search(line)
                if result:
                    self.version = result.group(1)

        self.output.info(f'Using version: {self.version}')

    def export(self):
        copy(self, "LICENSE.txt", src=self.recipe_folder, dst=self.export_folder)

    def export_sources(self):
        copy(self, "CMakeLists.txt", src=self.recipe_folder, dst=self.export_sources_folder)
        copy(self, "src/*", src=self.recipe_folder, dst=self.export_sources_folder)
        copy(self, "extras/*", src=self.recipe_folder, dst=self.export_sources_folder)
        copy(self, "CMake/*", src=self.recipe_folder, dst=self.export_sources_folder)

    def config_options(self):
        if self.settings.os == "Windows":
            del self.options.fPIC

    def configure(self):
        if self.options.shared:
            self.options.rm_safe("fPIC")

    def layout(self):
        cmake_layout(self)

    def validate(self):
        if self.settings.compiler.get_safe("cppstd"):
            check_min_cppstd(self, self._min_cppstd)
        # INFO: Conan 1.x does not specify cppstd by default, so we need to check the compiler version instead.
        minimum_version = self._compilers_minimum_version.get(str(self.settings.compiler), False)
        if minimum_version and Version(self.settings.compiler.version) < minimum_version:
            raise ConanInvalidConfiguration(f"{self.ref} requires C++{self._min_cppstd}, which your compiler doesn't support")

    def generate(self):
        tc = CMakeToolchain(self)
        tc.cache_variables["BUILD_TESTING"] = False
        tc.cache_variables["CATCH_INSTALL_DOCS"] = False
        tc.cache_variables["CATCH_INSTALL_EXTRAS"] = True
        tc.generate()

        deps = CMakeDeps(self)
        deps.generate()

    def build(self):
        cmake = CMake(self)
        cmake.configure()
        cmake.build()

    def package(self):
        copy(self, "LICENSE.txt", src=str(self.recipe_folder), dst=os.path.join(self.package_folder, "licenses"))
        cmake = CMake(self)
        cmake.install()
        rmdir(self, os.path.join(self.package_folder, "share"))
        rmdir(self, os.path.join(self.package_folder, "lib", "cmake"))
        copy(self, "*.cmake", src=os.path.join(self.export_sources_folder, "extras"),
                              dst=os.path.join(self.package_folder, "lib", "cmake", "Catch2"))

    def package_info(self):
        lib_suffix = "d" if self.settings.build_type == "Debug" else ""

        self.cpp_info.set_property("cmake_file_name", "Catch2")
        self.cpp_info.set_property("cmake_target_name", "Catch2::Catch2WithMain")
        self.cpp_info.set_property("pkg_config_name", "catch2-with-main")

        # Catch2
        self.cpp_info.components["catch2base"].set_property("cmake_file_name", "Catch2::Catch2")
        self.cpp_info.components["catch2base"].set_property("cmake_target_name", "Catch2::Catch2")
        self.cpp_info.components["catch2base"].set_property("pkg_config_name", "catch2")
        self.cpp_info.components["catch2base"].libs = ["Catch2" + lib_suffix]
        self.cpp_info.components["catch2base"].builddirs.append("lib/cmake/Catch2")

        # Catch2WithMain
        self.cpp_info.components["catch2main"].set_property("cmake_file_name", "Catch2::Catch2WithMain")
        self.cpp_info.components["catch2main"].set_property("cmake_target_name", "Catch2::Catch2WithMain")
        self.cpp_info.components["catch2main"].set_property("pkg_config_name", "catch2-with-main")
        self.cpp_info.components["catch2main"].libs = ["Catch2Main" + lib_suffix]
        self.cpp_info.components["catch2main"].requires = ["catch2base"]



#!/usr/bin/env python
#
# Copyright 2008, Google Inc.
# All rights reserved.
#
# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are
# met:
#
#     * Redistributions of source code must retain the above copyright
# notice, this list of conditions and the following disclaimer.
#     * Redistributions in binary form must reproduce the above
# copyright notice, this list of conditions and the following disclaimer
# in the documentation and/or other materials provided with the
# distribution.
#     * Neither the name of Google Inc. nor the names of its
# contributors may be used to endorse or promote products derived from
# this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
# "AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
# LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
# A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
# OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
# SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
# LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
# DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
# THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
# (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

r"""Tests the text output of Google C++ Mocking Framework.

To update the golden file:
gmock_output_test.py --build_dir=BUILD/DIR --gengolden
where BUILD/DIR contains the built gmock_output_test_ file.
gmock_output_test.py --gengolden
gmock_output_test.py

"""

from io import open    # pylint: disable=redefined-builtin, g-importing-member
import os
import re
import sys
from googlemock.test import gmock_test_utils


# The flag for generating the golden file
GENGOLDEN_FLAG = '--gengolden'

PROGRAM_PATH = gmock_test_utils.GetTestExecutablePath('gmock_output_test_')
COMMAND = [PROGRAM_PATH, '--gtest_stack_trace_depth=0', '--gtest_print_time=0']
GOLDEN_NAME = 'gmock_output_test_golden.txt'
GOLDEN_PATH = os.path.join(gmock_test_utils.GetSourceDir(), GOLDEN_NAME)


def ToUnixLineEnding(s):
  """Changes all Windows/Mac line endings in s to UNIX line endings."""

  return s.replace('\r\n', '\n').replace('\r', '\n')


def RemoveReportHeaderAndFooter(output):
  """Removes Google Test result report's header and footer from the output."""

  output = re.sub(r'.*gtest_main.*\n', '', output)
  output = re.sub(r'\[.*\d+ tests.*\n', '', output)
  output = re.sub(r'\[.* test environment .*\n', '', output)
  output = re.sub(r'\[=+\] \d+ tests .* ran.*', '', output)
  output = re.sub(r'.* FAILED TESTS\n', '', output)
  return output


def RemoveLocations(output):
  """Removes all file location info from a Google Test program's output.

  Args:
       output:  the output of a Google Test program.

  Returns:
       output with all file location info (in the form of
       'DIRECTORY/FILE_NAME:LINE_NUMBER: 'or
       'DIRECTORY\\FILE_NAME(LINE_NUMBER): ') replaced by
       'FILE:#: '.
  """

  return re.sub(r'.*[/\\](.+)(\:\d+|\(\d+\))\:', 'FILE:#:', output)


def NormalizeErrorMarker(output):
  """Normalizes the error marker, which is different on Windows vs on Linux."""

  return re.sub(r' error: ', ' Failure\n', output)


def RemoveMemoryAddresses(output):
  """Removes memory addresses from the test output."""

  return re.sub(r'@\w+', '@0x#', output)


def RemoveTestNamesOfLeakedMocks(output):
  """Removes the test names of leaked mock objects from the test output."""

  return re.sub(r'\(used in test .+\) ', '', output)


def GetLeakyTests(output):
  """Returns a list of test names that leak mock objects."""

  # findall() returns a list of all matches of the regex in output.
  # For example, if '(used in test FooTest.Bar)' is in output, the
  # list will contain 'FooTest.Bar'.
  return re.findall(r'\(used in test (.+)\)', output)


def GetNormalizedOutputAndLeakyTests(output):
  """Normalizes the output of gmock_output_test_.

  Args:
    output: The test output.

  Returns:
    A tuple (the normalized test output, the list of test names that have
    leaked mocks).
  """

  output = ToUnixLineEnding(output)
  output = RemoveReportHeaderAndFooter(output)
  output = NormalizeErrorMarker(output)
  output = RemoveLocations(output)
  output = RemoveMemoryAddresses(output)
  return (RemoveTestNamesOfLeakedMocks(output), GetLeakyTests(output))


def GetShellCommandOutput(cmd):
  """Runs a command in a sub-process, and returns its STDOUT in a string."""

  return gmock_test_utils.Subprocess(cmd, capture_stderr=False).output


def GetNormalizedCommandOutputAndLeakyTests(cmd):
  """Runs a command and returns its normalized output and a list of leaky tests.

  Args:
    cmd:  the shell command.
  """

  # Disables exception pop-ups on Windows.
  os.environ['GTEST_CATCH_EXCEPTIONS'] = '1'
  return GetNormalizedOutputAndLeakyTests(GetShellCommandOutput(cmd))


class GMockOutputTest(gmock_test_utils.TestCase):

  def testOutput(self):
    (output, leaky_tests) = GetNormalizedCommandOutputAndLeakyTests(COMMAND)
    golden_file = open(GOLDEN_PATH, 'rb')
    golden = golden_file.read().decode('utf-8')
    golden_file.close()
    # On Windows the repository might have been checked out with \r\n line
    # endings, so normalize it here.
    golden = ToUnixLineEnding(golden)

    # The normalized output should match the golden file.
    self.assertEqual(golden, output)

    # The raw output should contain 2 leaked mock object errors for
    # test GMockOutputTest.CatchesLeakedMocks.
    self.assertEqual(['GMockOutputTest.CatchesLeakedMocks',
                      'GMockOutputTest.CatchesLeakedMocks'],
                     leaky_tests)


if __name__ == '__main__':
  if sys.argv[1:] == [GENGOLDEN_FLAG]:
    (output, _) = GetNormalizedCommandOutputAndLeakyTests(COMMAND)
    golden_file = open(GOLDEN_PATH, 'wb')
    golden_file.write(output)
    golden_file.close()
    # Suppress the error "googletest was imported but a call to its main()
    # was never detected."
    os._exit(0)
  else:
    gmock_test_utils.Main()



#!/usr/bin/env python3
# pyright: strict

"""
stubgen.py: nanobind stub generation tool

This file provides both an API (``nanobind.stubgen.StubGen``) and a command
line interface to generate stubs for nanobind extensions.

To generate stubs on the command line, invoke the stub generator with a module
name, which will place the newly generated ``.pyi`` file directly into the
module folder.

```
python -m nanobind.stubgen <module name>
```

Specify ``-o <filename>`` or ``-O <path>`` to redirect the output somewhere
else in case this is not desired.

To programmatically generate stubs, construct an instance of the ``StubGen``
class and repeatedly call ``.put()`` to register modules or contents within the
modules (specific methods, classes, etc.). Afterwards, the ``.get()`` method
returns a string containing the stub declarations.

```
from nanobind.stubgen import StubGen
import my_module

sg = StubGen()
sg.put(my_module)
print(sg.get())
```

Internals:
----------

The implementation of stubgen performs a DFS traversal of the loaded module and
directly generates the stub in lockstep. There are no ASTs or other
intermediate data structures to keep things as simple as possible. Every kind
of object that could be encountered (functions, properties, values, types) has
a corresponding ``put_..`` method, along with the main ``put()`` entry point
that dispatches to the others as appropriate.

Internally, stub generation involves two potentially complex steps: converting
nanobind overload chains into '@overload' declarations that can be understood
by the 'typing' module, and turning default values back into Python
expressions. To make both steps well-defined, the implementation relies on an
internal ``__nb_signature__`` property that nanobind functions expose
specifically to simplify stub generation.

(Note that for now, the StubGen API is considered experimental and not subject
 to the semantic versioning policy used by the nanobind project.)
"""

import argparse
import builtins
import enum
from inspect import Signature, Parameter, signature, ismodule
import textwrap
import importlib
import importlib.machinery
import importlib.util
import types
import typing
from dataclasses import dataclass
from typing import Dict, Sequence, List, Optional, Tuple, cast, Generator, Any, Callable, Union, Protocol, Literal
from pathlib import Path
import re
import sys

if sys.version_info < (3, 9):
    from typing import Match, Pattern
else:
    from re import Match, Pattern

if sys.version_info < (3, 11):
    try:
        import typing_extensions
    except ImportError:
        raise RuntimeError(
            "stubgen.py requires the 'typing_extensions' package on Python <3.11"
        )
else:
    typing_extensions = None

SKIP_LIST = [
    # Various standard attributes found in modules, classes, etc.
    "__doc__", "__module__", "__name__", "__new__", "__builtins__",
    "__cached__", "__path__", "__version__", "__spec__", "__loader__",
    "__package__", "__nb_signature__", "__class_getitem__", "__orig_bases__",
    "__file__", "__dict__", "__weakref__", "__format__", "__nb_enum__",
    "__firstlineno__", "__static_attributes__", "__annotations__", "__annotate__",
    "__annotate_func__",

    # Auto-generated enum attributes. Type checkers synthesize these, so they
    # shouldn't appear in the stubs.
    "_new_member_", "_use_args_", "_member_names_", "_member_map_",
    "_value2member_map_", "_unhashable_values_", "_value_repr_",
]

# Interpreter-internal types.
TYPES_TYPES = {
    getattr(types, name): name for name in [
        "MethodDescriptorType",
        "MemberDescriptorType",
        "ModuleType",
    ]
}

# fmt: on

# This type is used to track per-module imports (``import name as desired_name``)
# during stub generation. The actual name in the stub is given by the value element.
# (name, desired_as_name) -> actual_as_name
ImportDict = Dict[Tuple[Optional[str], Optional[str]], Optional[str]]

# This type maps a module name to an `ImportDict` tuple that tracks the
# import declarations from that module.
# package_name -> ((name, desired_as_name) -> actual_as_name)
PackagesDict = Dict[str, ImportDict]

# Type of an entry of the ``__nb_signature__`` tuple of nanobind functions.
# It stores a function signature string, docstring, and a tuple of default function values.
# (signature_str, doc_str, (default_arg_1, default_arg_2, ...))
NbSignature = Tuple[Optional[str], Optional[str]]

# Type of an entry of the ``__nb_signature__`` tuple of nanobind functions.
NbFunctionSignature = Tuple[Optional[str], Optional[str], Optional[Tuple[Any, ...]]]

# Type of an entry of the ``__nb_signature__`` tuple of nanobind getters and setters.
NbGetterSetterSignature = Tuple[str, str]

class NamedObject(Protocol):
    """
    Typing protocol representing an object with __name__ and __module__ members
    """
    __module__: str
    __name__: str

class NbFunction(Protocol):
    """
    Typing protocol representing a nanobind function with its __nb_signature__ property
    """
    __module__: Literal["nanobind"]
    __name__: Literal["nb_func", "nb_method"]
    __nb_signature__: Tuple[NbFunctionSignature, ...]
    __call__: Callable[..., Any]


@typing.runtime_checkable
class NbGetterSetter(Protocol):
    __nb_signature__: Tuple[NbGetterSetterSignature, ...]


class NbStaticProperty(Protocol):
    """Typing protocol representing a nanobind static property"""
    __module__: Literal["nanobind"]
    __name__: Literal["nb_static_property"]
    fget: NbGetterSetter
    fset: NbGetterSetter


class NbType(Protocol):
    """typing protocol representing a nanobind type object"""
    __module__: Literal["nanobind"]
    __name__: Literal["nb_type"]
    __nb_signature__: str
    __bases__: Tuple[Any, ...]


@dataclass
class ReplacePattern:
    """
    A compiled query (regular expression) and replacement pattern. Patterns can
    be loaded using the ``load_pattern_file()`` function defined below
    """

    # A replacement patterns as produced by ``load_pattern_file()`` below
    query: Pattern[str]
    lines: List[str]
    matches: int


def create_subdirectory_for_module(module: types.ModuleType) -> bool:
    """
    When creating stubs recursively, prefer putting information directly
    into a ``submodule.pyi`` file unless the submodule has sub-submodules,
    or is defined in a nested directory (e.g. ``submodule/__init__.py``).
    In those two cases, put the stubs into ``submodule/__init__.pyi``
    """

    for child in module.__dict__.values():
        if ismodule(child):
            parent_name, _, _ = child.__name__.rpartition(".")
            if parent_name == module.__name__:
                return True

    return hasattr(module, '__file__') \
        and module.__file__ is not None \
        and module.__file__.endswith('__init__.py')


class StubGen:
    def __init__(
        self,
        module: types.ModuleType,
        recursive: bool = False,
        include_docstrings: bool = True,
        include_private: bool = False,
        include_internal_imports: bool = True,
        include_external_imports: bool = False,
        max_expr_length: int = 50,
        patterns: List[ReplacePattern] = [],
        quiet: bool = True,
        output_file: Optional[Path] = None
    ) -> None:
        # Module to check for name conflicts when adding helper imports
        self.module = module

        # Include docstrings in the generated stub?
        self.include_docstrings = include_docstrings

        # Recurse into submodules?
        self.recursive = recursive

        # Include private members that start or end with a single underscore?
        self.include_private = include_private

        # Include types and functions imported from the same package (but a different module)
        self.include_internal_imports = include_internal_imports

        # Include types and functions imported from external packages?
        self.include_external_imports = include_external_imports

        # Maximal length (in characters) before an expression gets abbreviated as '...'
        self.max_expr_length = max_expr_length

        # Replacement patterns as produced by ``load_pattern_file()`` below
        self.patterns = patterns

        # Set this to ``True`` if output to stdout is unacceptable
        self.quiet = quiet

        # Target filename, only needed for recursive stub generation
        self.output_file = output_file

        # ---------- Internal fields ----------

        # Current depth / indentation level
        self.depth = 0

        # Output will be appended to this string
        self.output = ""

        # A stack to avoid infinite recursion
        self.stack: List[object] = []

        # An identifier associated with the top element of the stack
        self.prefix = module.__name__

        # Dictionary to keep track of import directives added by the stub generator
        # Maps package_name -> ((name, desired_as_name) -> actual_as_name)
        self.imports: PackagesDict = {}

        # ---------- Regular expressions ----------

        # Negative lookbehind matching word boundaries except '.'
        sep_before = r"(?<![\\B\.])"

        # Negative lookforward matching word boundaries except '.'
        sep_after = r"(?![\\B\.])"

        # Regexp matching a Python identifier
        identifier = r"[^\d\W]\w*"

        # Precompile RE a sequence of identifiers separated by periods
        self.id_seq = re.compile(
            sep_before
            + "((?:"
            + identifier
            + r"\.)+)("
            + identifier
            + r")\b"
            + sep_after
        )

        # Precompile RE to extract nanobind nd-arrays
        self.ndarray_re = re.compile(
            sep_before + r"(numpy\.ndarray|ndarray|torch\.Tensor)\[([^\]]*)\]"
        )

        # Types which moved from typing.* to collections.abc in Python 3.9
        self.abc_re = re.compile(
            'typing.(AsyncGenerator|AsyncIterable|AsyncIterator|Awaitable|Callable|'
            'Collection|Container|Coroutine|Generator|Hashable|ItemsView|'
            'Iterable|Iterator|KeysView|Mapping|MappingView|MutableMapping|'
            'MutableSequence|MutableSet|Sequence|ValuesView)'
        )

    def write(self, s: str) -> None:
        """Append raw characters to the output"""
        self.output += s

    def write_ln(self, line: str) -> None:
        """Append an indented line"""
        if len(line) != 0 and not line.isspace():
            self.output += "    " * self.depth + line
        self.output += "\n"

    def write_par(self, line: str) -> None:
        """Append an indented paragraph"""
        self.output += textwrap.indent(line, "    " * self.depth)

    def put_docstr(self, docstr: str) -> None:
        """Append an indented single or multi-line docstring"""
        docstr = textwrap.dedent(docstr).strip()
        raw_str = ""
        if "''" in docstr or "\\" in docstr:
            # Escape all double quotes so that no unquoted triple quote can exist
            docstr = docstr.replace("''", "\\'\\'")
            raw_str = "r"
        if len(docstr) > 70 or "\n" in docstr:
            docstr = "\n" + docstr + "\n"
        docstr = f'{raw_str}"""{docstr}"""\n'
        self.write_par(docstr)

    def put_nb_overload(self, fn: NbFunction, sig: NbFunctionSignature, name: Optional[str] = None) -> None:
        """
        The ``put_nb_func()`` repeatedly calls this method to render the
        individual method overloads.
        """
        sig_str, docstr, start = cast(str, sig[0]), cast(str, sig[1]), 0

        # Label anonymous functions
        if sig_str.startswith("def (") and name is not None:
            sig_str = "def " + name + sig_str[4:]

        # Simplify type names present in the signature
        paren = sig_str.find("(")
        sig_str = sig_str[:paren] + self.simplify_types(sig_str[paren:])

        # Substitute in string versions of the default arguments
        default_args = sig[2]
        if default_args:
            for index, arg in enumerate(default_args):
                pos = -1
                pattern = None
                arg_str = None

                # First, handle the case where the user overrode the default value signature
                if isinstance(arg, str):
                    pattern = f"\\={index}"
                    pos = sig_str.find(pattern, start)
                    if pos >= 0:
                        arg_str = arg

                # General case
                if pos < 0:
                    pattern = f"\\{index}"
                    pos = sig_str.find(pattern, start)

                if pos < 0:
                    raise Exception(
                        "Could not locate default argument in function signature"
                    )

                if not arg_str:
                    # Call expr_str to convert the default value to a string.
                    # Abbreviate with '...' if it is too long.
                    expr = self.expr_str(arg, abbrev=True)
                    arg_str = expr if expr else "..."

                assert (
                    "\n" not in arg_str
                ), "Default argument string may not contain newlines."

                assert pattern is not None
                sig_str = sig_str[:pos] + arg_str + sig_str[pos + len(pattern) :]
                start = pos + len(arg_str)

        if type(fn).__name__ == "nb_func" and self.depth > 0:
            self.write_ln("@staticmethod")

        if not docstr or not self.include_docstrings:
            for s in sig_str.split("\n"):
                self.write_ln(s)
            self.output = self.output[:-1] + ": ...\n"
        else:
            docstr = textwrap.dedent(docstr)
            for s in sig_str.split("\n"):
                self.write_ln(s)
            self.output = self.output[:-1] + ":\n"
            self.depth += 1
            self.put_docstr(docstr)
            self.depth -= 1
        self.write("\n")

    def put_nb_func(self, fn: NbFunction, name: Optional[str] = None) -> None:
        """Append a nanobind function binding to the stub"""
        sigs = fn.__nb_signature__
        count = len(sigs)
        assert count > 0
        if count == 1:
            # No overloads write directly
            self.put_nb_overload(fn, sigs[0], name)
        else:
            # Render an @overload-decorated chain
            overload = self.import_object("typing", "overload")
            for s in sigs:
                self.write_ln(f"@{overload}")
                self.put_nb_overload(fn, s, name)

    def put_function(self, fn: Callable[..., Any], name: Optional[str] = None, parent: Optional[object] = None):
        """Append a function of an arbitrary type to the stub"""
        # Don't generate a constructor for nanobind classes that aren't constructible
        if name == "__init__" and type(parent).__name__.startswith("nb_type"):
            return

        fn_module = getattr(fn, "__module__", None)
        fn_name = getattr(fn, "__name__", None)

        # Check if this function is an alias from *another* module
        if name and fn_module and fn_module != self.module.__name__:
            self.put_value(fn, name)
            return

        # Check if this function is an alias from the *same* module
        if name and fn_name and name != fn_name:
            self.write_ln(f"{name} = {fn_name}\n")
            return

        # Special handling for nanobind functions with overloads
        if type(fn).__module__ == "nanobind":
            fn = cast(NbFunction, fn)
            self.put_nb_func(fn, name)
            return

        if isinstance(fn, staticmethod):
            self.write_ln("@staticmethod")
            fn = fn.__func__
        elif isinstance(fn, classmethod):
            self.write_ln("@classmethod")
            fn = fn.__func__

        if name is None:
            name = fn.__name__
            assert name

        overloads: Sequence[Callable[..., Any]] = []
        if hasattr(fn, "__module__"):
            if typing_extensions:
                overloads = typing_extensions.get_overloads(fn)
            else:
                overloads = typing.get_overloads(fn)

        if not overloads:
            overloads = [fn]

        for i, fno in enumerate(overloads):
            if len(overloads) > 1:
                overload = self.import_object("typing", "overload")
                self.write_ln(f"@{overload}")

            try:
                sig = signature(fno)
            except ValueError:
                sig = None

            if sig is not None:
                sig_str = f"{name}{self.signature_str(sig)}"
            else:
                # If inspect.signature fails, use a maximally permissive type.
                any_type = self.import_object("typing", "Any")
                sig_str = f"{name}(*args, **kwargs) -> {any_type}"

            # Potentially copy docstring from the implementation function
            docstr = fno.__doc__
            if i == 0 and not docstr and fn.__doc__:
                docstr = fn.__doc__

            if not docstr or not self.include_docstrings:
                self.write_ln("def " + sig_str + ": ...")
            else:
                self.write_ln("def " + sig_str + ":")
                self.depth += 1
                self.put_docstr(docstr)
                self.depth -= 1
            self.write("\n")

    def put_property(self, prop: property, name: Optional[str]):
        """Append a Python 'property' object"""
        fget, fset = prop.fget, prop.fset
        self.write_ln("@property")
        self.put(fget, name=name)
        if fset:
            self.write_ln(f"@{name}.setter")
            docstrings_backup = self.include_docstrings
            if isinstance(fget, NbGetterSetter) and isinstance(fset, NbGetterSetter):
                doc1 = fget.__nb_signature__[0][1]
                doc2 = fset.__nb_signature__[0][1]
                if doc1 and doc2 and doc1 == doc2:
                    self.include_docstrings = False
            self.put(prop.fset, name=name)
            self.include_docstrings = docstrings_backup

    def put_nb_static_property(self, name: Optional[str], prop: NbStaticProperty):
        """Append a 'nb_static_property' object"""
        getter_sig = prop.fget.__nb_signature__[0][0]
        pos = getter_sig.find("/) -> ")
        if pos == -1:
            raise RuntimeError(f"Static property '{name}' ({getter_sig}) has an invalid signature!")
        getter_sig = getter_sig[pos + 6 :]
        self.write_ln(f"{name}: {getter_sig} = ...")
        if prop.__doc__ and self.include_docstrings:
            self.put_docstr(prop.__doc__)
        self.write("\n")

    def put_type(self, tp: NbType, name: Optional[str]):
        """Append a 'nb_type' type object"""
        tp_name, tp_mod_name = tp.__name__, tp.__module__
        mod_name = self.module.__name__

        if name and (name != tp_name or mod_name != tp_mod_name):
            same_module = tp_mod_name == mod_name
            same_toplevel_module = tp_mod_name.split(".")[0] == mod_name.split(".")[0]

            if same_module:
                # This is an alias of a type in the same module or same top-level module
                if sys.version_info >= (3, 10, 0):
                    alias_tp = self.import_object("typing", "TypeAlias")
                else:
                    alias_tp = self.import_object("typing_extensions", "TypeAlias")
                self.write_ln(f"{name}: {alias_tp} = {tp.__qualname__}\n")
            elif self.include_external_imports or (same_toplevel_module and self.include_internal_imports):
                # Import from a different module
                self.put_value(tp, name)
        else:
            docstr = tp.__doc__
            tp_dict = dict(tp.__dict__)
            tp_bases: Union[List[str], Tuple[Any, ...], None] = None

            if "__nb_signature__" in tp.__dict__:
                # Types with a custom signature override
                for s in tp.__nb_signature__.split("\n"):
                    self.write_ln(self.simplify_types(s))
                self.output = self.output[:-1] + ":\n"
            else:
                self.write_ln(f"class {tp_name}:")
                if tp_bases is None:
                    tp_bases = getattr(tp, "__orig_bases__", None)
                    if tp_bases is None:
                        tp_bases = tp.__bases__
                    tp_bases = [self.type_str(base) for base in tp_bases]

                if tp_bases != ["object"]:
                    self.output = self.output[:-2] + "("
                    for i, base in enumerate(tp_bases):
                        if i:
                            self.write(", ")
                        self.write(base)
                    self.write("):\n")

            self.depth += 1
            output_len = len(self.output)
            if docstr and self.include_docstrings:
                self.put_docstr(docstr)
                if len(tp_dict):
                    self.write("\n")
            for k, v in tp_dict.items():
                self.put(v, k, tp)
            if output_len == len(self.output):
                self.write_ln("pass\n")
            self.depth -= 1

    def is_function(self, tp: type) -> bool:
        """
        Test if this is one of the many types of built-in functions supported
        by Python, or if it is a nanobind ``nb_func``.
        """
        return (
            issubclass(tp, types.FunctionType)
            or issubclass(tp, types.BuiltinFunctionType)
            or issubclass(tp, types.BuiltinMethodType)
            or issubclass(tp, types.WrapperDescriptorType)
            or issubclass(tp, staticmethod)
            or issubclass(tp, classmethod)
            or (tp.__module__ == "nanobind" and tp.__name__ == "nb_func")
        )

    def put_value(self, value: object, name: str, parent: Optional[object] = None, abbrev: bool = True) -> None:
        """
        Render a ``name: type = value`` assignment at the module, class, or
        enum scope.

        The parameter ``abbrev`` indicates if it is acceptable to reduce very
        long expressions to ``...``.
        """
        tp = type(value)

        # Ignore module imports of non-type values like 'from typing import Optional'
        if (
            not self.include_external_imports
            and tp.__module__ == "typing"
            and str(value) == f"typing.{name}"
        ):
            return

        if tp.__module__ == '__future__':
            return

        if isinstance(parent, type) and issubclass(tp, parent):
            # This is an entry of an enumeration
            self.write_ln(f"{name} = {typing.cast(enum.Enum, value).value}")
            if value.__doc__ and self.include_docstrings:
                self.put_docstr(value.__doc__)
            self.write("\n")
        elif self.is_function(tp) or isinstance(value, type):
            named_value = cast(NamedObject, value)
            same_toplevel_module = named_value.__module__.split(".")[0] == self.module.__name__.split(".")[0]

            if self.include_external_imports or (same_toplevel_module and self.include_internal_imports):
                # This is a function or a type, import it from its actual source
                self.import_object(named_value.__module__, named_value.__name__, name)
        else:
            value_str = self.expr_str(value, abbrev)

            if value_str is None:
                value_str = "..."

            # Catch a few different typing.* constructs
            if self.is_type_var(tp):
                types = ""
            elif typing.get_origin(value):
                if sys.version_info >= (3, 10, 0):
                    types = ": " + self.import_object("typing", "TypeAlias")
                else:
                    types = ": " + self.import_object("typing_extensions", "TypeAlias")
            else:
                types = f": {self.type_str(tp)}"

            self.write_ln(f"{name}{types} = {value_str}\n")

    def is_type_var(self, tp: type) -> bool:
        if issubclass(tp, typing.TypeVar):
            return True
        if sys.version_info >= (3, 10) and issubclass(tp, typing.ParamSpec):
            return True
        if sys.version_info >= (3, 11) and issubclass(tp, typing.TypeVarTuple):
            return True
        if typing_extensions is not None:
            if issubclass(
                tp,
                (
                    typing_extensions.TypeVar,
                    typing_extensions.ParamSpec,
                    typing_extensions.TypeVarTuple
                )
            ):
                return True
        return False


    def simplify_types(self, s: str) -> str:
        """
        Process types that occur within a signature string ``s`` and simplify
        them. This function applies the following rules:

        - "local_module.X" -> "X"

        - "other_module.X" -> "other_module.X"
          (with "import other_module" added at top)

        - "builtins.X" -> "X"

        - "NoneType" -> "None"

        - "ndarray[...]" -> "Annotated[NDArray[dtype], dict(..extras..)]"

        - "collections.abc.X" -> "X"
          (with "from collections.abc import X" added at top)

        - "typing.X" -> "X"
          (with "from typing import X" added at top, potentially
           changed to 'collections.abc' on newer Python versions)
        """

        # Process nd-array type annotations so that MyPy accepts them
        s = self.ndarray_re.sub(lambda m: self._format_ndarray(m.group(2)), s)

        if sys.version_info >= (3, 9, 0):
            s = self.abc_re.sub(r'collections.abc.\1', s)

        # Process other type names and add suitable import statements
        def process_general(m: Match[str]) -> str:
            def is_valid_module(module_name: str) -> bool:
                try:
                    importlib.util.find_spec(module_name)
                    # If we get here, the module exists and has a valid spec.
                    return True
                except ValueError:
                    # The module exists but has no spec, `find_spec` raises a
                    # `ValueError`, so if we get here, the module does exist.
                    return True
                except ModuleNotFoundError:
                    return False

            full_name, mod_name, cls_name = m.group(0), m.group(1)[:-1], m.group(2)

            if mod_name == "builtins":
                # Simplify builtins
                return cls_name if cls_name != "NoneType" else "None"
            if full_name.startswith(self.module.__name__):
                # Strip away the module prefix for local classes
                return full_name[len(self.module.__name__) + 1 :]
            elif mod_name == "typing" or mod_name == "collections.abc":
                # Import frequently-occurring typing classes and ABCs directly
                return self.import_object(mod_name, cls_name)
            else:
                # Handle nested names. While mod_name isn't a valid module, then
                # move the last segment of the name from mod_name to cls_name
                # and try again until we have the right partition.
                search_mod_name = mod_name
                search_cls_name = cls_name
                while search_mod_name:
                    if is_valid_module(search_mod_name):
                        mod_name = search_mod_name
                        cls_name = search_cls_name
                        break
                    search_mod_name, _, symbol = search_mod_name.rpartition(".")
                    search_cls_name = f"{symbol}.{search_cls_name}"

                # Import the module and reference the contained class by name
                self.import_object(mod_name, None)
                return full_name

        s = self.id_seq.sub(process_general, s)

        return s

    def _format_ndarray(self, annotation: str) -> str:
        """Improve NumPy type annotations for static type checking"""
        dtype = None
        m = re.search(r"dtype=(\w+)", annotation)

        if m:
            dtype = "numpy."+ m.group(1)
            dtype = dtype.replace('bool', 'bool_')
            annotation = re.sub(r"dtype=\w+,?\s*", "", annotation).rstrip(", ")

        # Turn shape notation into a valid Python type expression
        annotation = annotation.replace("*", "None").replace("(None)", "(None,)")

        # Build type while potentially preserving extra information as an annotation
        ndarray = self.import_object("numpy.typing", "NDArray")
        result = f"{ndarray}[{dtype}]" if dtype else ndarray
        if annotation:
            annotated = self.import_object("typing", "Annotated")
            result = f"{annotated}[{result}, dict({annotation})]"

        return result

    def apply_pattern(self, query: str, value: object) -> bool:
        """
        Check if ``value`` matches an entry of a pattern file. Applies the
        pattern and returns ``True`` in that case, otherwise returns ``False``.
        """

        match: Optional[Match[str]] = None
        pattern: Optional[ReplacePattern] = None

        for pattern in self.patterns:
            match = pattern.query.search(query)
            if match:
                break

        if not match or not pattern:
            return False

        for line in pattern.lines:
            ls = line.strip()
            if ls == "\\doc":
                # Docstring reference
                tp = type(value)
                doc: Optional[str] = None
                if tp.__module__ == "nanobind" and tp.__name__ in (
                    "nb_func",
                    "nb_method",
                ):
                    value = cast(NbFunction, value)
                    for tp_i in value.__nb_signature__:
                        doc = tp_i[1]
                        if doc:
                            break
                else:
                    doc = getattr(value, "__doc__", None)
                self.depth += 1
                if doc and self.include_docstrings:
                    self.put_docstr(doc)
                else:
                    self.write_ln("pass")
                self.depth -= 1
                continue
            elif ls.startswith("\\from "):
                items = ls[5:].split(" import ")
                if len(items) != 2:
                    raise RuntimeError(f"Could not parse import declaration {ls}")
                for item in items[1].strip("()").split(","):
                    item_list = item.split(" as ")
                    import_module, import_name = (
                        items[0].strip(),
                        item_list[0].strip(),
                    )
                    import_as = item_list[1].strip() if len(item_list) > 1 else None
                    self.import_object(import_module, import_name, import_as)
                continue

            groups = match.groups()
            for i in reversed(range(len(groups))):
                line = line.replace(f"\\{i+1}", groups[i])
            for k, v in match.groupdict().items():
                line = line.replace(f"\\{k}", v)
            self.write_ln(line)

        # Success, pattern was applied
        return True

    def put(self, value: object, name: Optional[str] = None, parent: Optional[object] = None) -> None:
        old_prefix = self.prefix

        if value in self.stack:
            # Avoid infinite recursion due to cycles
            return

        try:
            self.stack.append(value)
            self.prefix = self.prefix + (("." + name) if name else "")

            # Check if an entry in a provided pattern file matches
            if self.apply_pattern(self.prefix, value):
                return

            # Exclude various standard elements found in modules, classes, etc.
            if name in SKIP_LIST:
                return

            is_type_alias = typing.get_origin(value) or (
                isinstance(value, type)
                and (value.__name__ != name or value.__module__ != self.module.__name__)
            )

            # Ignore private members unless the user requests their inclusion
            if (
                not self.include_private
                and name
                and not is_type_alias
                and len(name) > 2
                and (
                    (name[0] == "_" and name[1] != "_")
                    or (name[-1] == "_" and name[-2] != "_")
                )
            ):
                return

            tp = type(value)
            tp_mod, tp_name = tp.__module__, tp.__name__

            if ismodule(value):
                if len(self.stack) != 1:
                    value_name_s = value.__name__.split(".")
                    module_name_s = self.module.__name__.split(".")
                    is_external = value_name_s[0] != module_name_s[0]
                    if not self.include_external_imports and is_external:
                        return

                    # Do not include submodules in the same stub, but include a directive to import them
                    self.import_object(value.__name__, name=None, as_name=name)

                    # If the user requested this, generate recursive stub files as well
                    if self.recursive and value_name_s[:-1] == module_name_s and self.output_file:
                        if create_subdirectory_for_module(value):
                            # Create a new subdirectory and start with an __init__.pyi file there
                            dir_name = self.output_file.parents[0] / value_name_s[-1]
                            dir_name.mkdir(parents=False, exist_ok=True)
                            output_file = dir_name / '__init__.pyi'
                        else:
                            output_file = self.output_file.parents[0] / (value_name_s[-1] + '.pyi')

                        sg = StubGen(
                            module=value,
                            recursive=self.recursive,
                            include_docstrings=self.include_docstrings,
                            include_private=self.include_private,
                            include_external_imports=self.include_external_imports,
                            include_internal_imports=self.include_internal_imports,
                            max_expr_length=self.max_expr_length,
                            patterns=self.patterns,
                            output_file=output_file,
                            quiet=self.quiet
                        )

                        sg.put(value)
                        output_file = output_file.resolve()

                        if not self.quiet:
                            print(f'  - writing stub "{output_file}" ..')

                        with open(output_file, "w", encoding='utf-8') as f:
                            f.write(sg.get())
                    return
                else:
                    self.apply_pattern(self.prefix + ".__prefix__", None)
                    # using value.__dict__ rather than inspect.getmembers
                    # to preserve insertion order
                    for name, child in value.__dict__.items():
                        self.put(child, name=name, parent=value)
                    self.apply_pattern(self.prefix + ".__suffix__", None)
            elif self.is_function(tp):
                value = cast(NbFunction, value)
                self.put_function(value, name, parent)
            elif issubclass(tp, type):
                value = cast(NbType, value)
                self.put_type(value, name)
            elif tp_mod == "nanobind":
                if tp_name == "nb_method":
                    value = cast(NbFunction, value)
                    self.put_function(value, name)
                elif tp_name == "nb_static_property":
                    value = cast(NbStaticProperty, value)
                    self.put_nb_static_property(name, value)
            elif tp_mod == "builtins":
                if tp is property:
                    value = cast(property, value)
                    self.put_property(value, name)
                else:
                    assert name is not None
                    abbrev = name != "__all__"
                    self.put_value(value, name, parent, abbrev=abbrev)
            else:
                assert name is not None
                self.put_value(value, name, parent)
        finally:
            self.stack.pop()
            self.prefix = old_prefix

    def import_object(
        self, module: str, name: Optional[str], as_name: Optional[str] = None
    ) -> str:
        """
        Import a type (e.g. typing.Optional) used within the stub, ensuring
        that this does not cause conflicts. Specify ``as_name`` to ensure that
        the import is bound to a specified name.

        When ``name`` is None, the entire module is imported.
        """
        if module == "builtins" and name and (not as_name or name == as_name):
            return name

        # Rewrite module name if this is relative import from a submodule
        if module.startswith(self.module.__name__) and module != self.module.__name__:
            module_short = module[len(self.module.__name__) :]
            if not name and as_name and module_short[0] == ".":
                name = as_name = module_short[1:]
                module_short = "."
        else:
            module_short = module

        # Query a cache of previously imported objects
        imports_module: Optional[ImportDict] = self.imports.get(module_short, None)
        if not imports_module:
            imports_module = {}
            self.imports[module_short] = imports_module

        key = (name, as_name)
        final_name = imports_module.get(key, None)
        if final_name:
            return final_name

        # Cache miss, import the object
        final_name = as_name if as_name else name

        # If no as_name constraint was set, potentially adjust the name to
        # avoid conflicts with an existing object of the same name
        if name and not as_name:
            test_name = name
            while True:
                # Accept the name if there are no conflicts
                if not hasattr(self.module, test_name):
                    break
                value = getattr(self.module, test_name)
                try:
                    if module == ".":
                        mod_o = self.module
                    else:
                        mod_o = importlib.import_module(module)

                    # If there is a conflict, accept it if it refers to the same object
                    if getattr(mod_o, name) is value:
                        break
                except ImportError:
                    pass

                # Prefix with an underscore
                test_name = "_" + test_name
            final_name = test_name

        imports_module[key] = final_name
        return final_name if final_name else ""

    def expr_str(self, e: Any, abbrev: bool = True) -> Optional[str]:
        """
        Attempt to convert a value into valid Python syntax that regenerates
        that value. When ``abbrev`` is True, the implementation gives up and
        returns ``None`` when the expression is considered to be too
        complicated.
        """
        tp = type(e)
        if issubclass(tp, (bool, int, type(None), type(builtins.Ellipsis))):
            s = repr(e)
            if len(s) < self.max_expr_length or not abbrev:
                return s
        elif issubclass(tp, float):
            s = repr(e)
            if "inf" in s or "nan" in s:
                s = f"float('{s}')"
            if len(s) < self.max_expr_length or not abbrev:
                return s
        elif issubclass(tp, type) or typing.get_origin(e):
            return self.type_str(e)
        elif issubclass(tp, typing.ForwardRef):
            return f'"{e.__forward_arg__}"'
        elif issubclass(tp, enum.Enum):
            return self.type_str(tp) + '.' + e.name
        elif (sys.version_info >= (3, 10) and issubclass(tp, typing.ParamSpec)) \
            or (typing_extensions is not None and issubclass(tp, typing_extensions.ParamSpec)):
            tv = self.import_object(tp.__module__, "ParamSpec")
            return f'{tv}("{e.__name__}")'
        elif (sys.version_info >= (3, 11) and issubclass(tp, typing.TypeVarTuple)) \
            or (typing_extensions is not None and issubclass(tp, typing_extensions.TypeVarTuple)):
            tv = self.import_object(tp.__module__, "TypeVarTuple")
            return f'{tv}("{e.__name__}")'
        elif issubclass(tp, typing.TypeVar):
            tv = self.import_object("typing", "TypeVar")
            s = f'{tv}("{e.__name__}"'
            for v in getattr(e, "__constraints__", ()):
                v = self.type_str(v)
                assert v
                s += ", " + v
            if v := getattr(e, "__bound__", None):
                v = self.type_str(v)
                assert v
                s += ", bound=" + v
            for k in ["contravariant", "covariant", "infer_variance"]:
                v = getattr(e, f"__{k}__", None)
                if v:
                    v = self.expr_str(v, abbrev=False)
                    if v is None:
                        return None
                    s += f", {k}=" + v
            s += ")"
            return s
        elif issubclass(tp, str):
            s = repr(e)
            if len(s) < self.max_expr_length or not abbrev:
                return s
        elif issubclass(tp, list) or issubclass(tp, tuple):
            e = [self.expr_str(v, abbrev) for v in e]
            if None in e:
                return None
            if issubclass(tp, list):
                s = "[" + ", ".join(e) + "]"
            else:
                s = "(" + ", ".join(e) + ")"
            if len(s) < self.max_expr_length or not abbrev:
                return s
        elif issubclass(tp, dict):
            e = [
                (self.expr_str(k, abbrev), self.expr_str(v, abbrev))
                for k, v in e.items()
            ]
            s = "{"
            for i, (k, v) in enumerate(e):
                if k is None or v is None:
                    return None
                s += k + " : " + v
                if i + 1 < len(e):
                    s += ", "
            s += "}"
            if len(s) < self.max_expr_length or not abbrev:
                return s
            pass
        return None

    def signature_str(self, s: Signature) -> str:
        """Convert an inspect.Signature to into valid Python syntax"""
        posonly_sep, kwonly_sep = False, True
        params: List[str] = []

        # Logic for placing '*' and '/' based on the
        # signature.Signature implementation
        for param in s.parameters.values():
            kind = param.kind

            if kind == Parameter.POSITIONAL_ONLY:
                posonly_sep = True
            elif posonly_sep:
                params.append("/")
                posonly_sep = False

            if kind == Parameter.VAR_POSITIONAL:
                kwonly_sep = False
            elif kind == Parameter.KEYWORD_ONLY and kwonly_sep:
                params.append("*")
                kwonly_sep = False
            params.append(self.param_str(param))

        if posonly_sep:
            params.append("/")

        result = f"({', '.join(params)})"
        if s.return_annotation != Signature.empty:
            result += " -> " + self.type_str(s.return_annotation)
        return result

    def param_str(self, p: Parameter) -> str:
        result = ""
        if p.kind == Parameter.VAR_POSITIONAL:
            result += "*"
        elif p.kind == Parameter.VAR_KEYWORD:
            result += "**"
        result += p.name
        has_type = p.annotation != Parameter.empty
        has_def = p.default != Parameter.empty

        if has_type:
            result += ": " + self.type_str(p.annotation)
        if has_def:
            result += " = " if has_type else "="
            p_default_str = self.expr_str(p.default)
            if p_default_str is None:
                # self.expr_str(p.default) could return None in some rare cases,
                # e.g. p.default is a nanobind object. If so, use ellipsis as a placeholder.
                p_default_str = "..."
            assert p_default_str
            result += p_default_str
        return result

    def type_str(self, tp: Union[List[Any], Tuple[Any, ...], Dict[Any, Any], Any]) -> str:
        """Attempt to convert a type into a Python expression which reproduces it"""
        origin, args = typing.get_origin(tp), typing.get_args(tp)

        if isinstance(tp, str):
            result = tp
        elif isinstance(tp, typing.TypeVar):
            return tp.__name__
        elif isinstance(tp, typing.ForwardRef):
            return repr(tp.__forward_arg__)
        elif isinstance(tp, list):
            list_gen: Generator[str, None, None] = (self.type_str(a) for a in tp)
            return "[" + ", ".join(list_gen) + "]"
        elif isinstance(tp, tuple):
            tuple_gen: Generator[str, None, None] = (self.type_str(a) for a in tp)
            return "(" + ", ".join(tuple_gen) + ")"
        elif isinstance(tp, dict):
            dict_gen: Generator[str, None, None] = (repr(k) + ": " + self.type_str(v) for k, v in tp.items())
            return (
                "{"
                + ", ".join(dict_gen)
                + "}"
            )
        elif origin and args:
            args_gen: Generator[str, None, None] = (self.type_str(a) for a in args)
            result = (
                self.type_str(origin)
                + "["
                + ", ".join(args_gen)
                + "]"
            )
        elif tp in TYPES_TYPES:
            result = f"types.{TYPES_TYPES[tp]}"
        elif tp is Ellipsis:
            result = "..."
        elif isinstance(tp, type):
            result = tp.__module__ + "." + tp.__qualname__
        else:
            result = repr(tp)
        return self.simplify_types(result)

    def check_party(self, module: str) -> Literal[0, 1, 2]:
        """
        Check source of module
        0 = From stdlib
        1 = From 3rd party package
        2 = From the package being built
        """
        if module.startswith(".") or module.split('.')[0] == self.module.__name__.split('.')[0]:
            return 2

        try:
            spec = importlib.util.find_spec(module)
        except (ModuleNotFoundError, ValueError):
            return 1

        if spec:
            if spec.origin and ("site-packages" in spec.origin or "dist-packages" in spec.origin):
                return 1
            else:
                return 0
        else:
            return 1

    def get(self) -> str:
        """Generate the final stub output"""
        s = ""
        last_party = None

        for module in sorted(self.imports, key=lambda i: str(self.check_party(i)) + i):
            imports = self.imports[module]
            items: List[str] = []
            party = self.check_party(module)

            if party != last_party:
                if last_party is not None:
                    s += "\n"
                last_party = party

            for (k, v1), v2 in imports.items():
                if k is None:
                    if v1 and v1 != module:
                        s += f"import {module} as {v1}\n"
                    elif v1 is None or (k, None) not in imports:
                        s += f"import {module}\n"
                else:
                    if k != v2 or v1:
                        items.append(f"{k} as {v2}")
                    else:
                        items.append(k)

            items = sorted(items)
            if items:
                items_v0 = ", ".join(items)
                items_v0 = f"from {module} import {items_v0}\n"
                items_v1 = "(\n    " + ",\n    ".join(items) + "\n)"
                items_v1 = f"from {module} import {items_v1}\n"
                s += items_v0 if len(items_v0) <= 70 else items_v1

        s += "\n\n"

        # Append the main generated stub
        s += self.output

        return s.rstrip() + "\n"

def parse_options(args: List[str]) -> argparse.Namespace:
    parser = argparse.ArgumentParser(
        prog="python -m nanobind.stubgen",
        description="Generate stubs for nanobind-based extensions.",
    )

    parser.add_argument(
        "-o",
        "--output-file",
        metavar="FILE",
        dest="output_file",
        default=None,
        help="write generated stubs to the specified file",
    )

    parser.add_argument(
        "-O",
        "--output-dir",
        metavar="PATH",
        dest="output_dir",
        default=None,
        help="write generated stubs to the specified directory",
    )

    parser.add_argument(
        "-i",
        "--import",
        action="append",
        metavar="PATH",
        dest="imports",
        default=[],
        help="add the directory to the Python import path (can specify multiple times)",
    )

    parser.add_argument(
        "-m",
        "--module",
        action="append",
        metavar="MODULE",
        dest="modules",
        default=[],
        help="generate a stub for the specified module (can specify multiple times)",
    )

    parser.add_argument(
        "-r",
        "--recursive",
        default=False,
        action="store_true",
        dest="recursive",
        help="recursively process submodules",
    )

    parser.add_argument(
        "-M",
        "--marker-file",
        action="append",
        metavar="FILE",
        dest="marker_file",
        default=[],
        help="generate a marker file (usually named 'py.typed', can specify multiple times)",
    )

    parser.add_argument(
        "-p",
        "--pattern-file",
        metavar="FILE",
        dest="pattern_file",
        default=None,
        help="apply the given patterns to the generated stub (see the docs for syntax)",
    )

    parser.add_argument(
        "-P",
        "--include-private",
        dest="include_private",
        default=False,
        action="store_true",
        help="include private members (with single leading or trailing underscore)",
    )

    parser.add_argument(
        "-D",
        "--exclude-docstrings",
        dest="include_docstrings",
        default=True,
        action="store_false",
        help="exclude docstrings from the generated stub",
    )

    parser.add_argument(
        "--exclude-values",
        dest="exclude_values",
        default=False,
        action="store_true",
        help="force the use of ... for values",
    )

    parser.add_argument(
        "-q",
        "--quiet",
        default=False,
        action="store_true",
        help="do not generate any output in the absence of failures",
    )

    opt = parser.parse_args(args)
    if len(opt.modules) == 0:
        parser.error("At least one module must be specified.")
    if len(opt.modules) > 1 and opt.output_file:
        parser.error(
            "The -o option can only be specified when a single module is being processed."
        )
    if opt.recursive and opt.output_file:
        parser.error(
            "The -o option is not compatible with recursive stub generation (-r)."
        )
    return opt


def load_pattern_file(fname: str) -> List[ReplacePattern]:
    """
    Load a pattern file from disk and return a list of pattern instances that
    includes precompiled versions of all of the contained regular expressions.
    """

    with open(fname, "r", encoding='utf-8') as f:
        f_lines = f.readlines()

    patterns: List[ReplacePattern] = []

    def add_pattern(query: str, lines: List[str]):
        # Exactly 1 empty line at the end
        while lines and (lines[-1].isspace() or len(lines[-1]) == 0):
            lines.pop()
        lines.append("")

        # Identify deletions (replacement by only whitespace)
        if all((p.isspace() or len(p) == 0 for p in lines)):
            lines = []
        patterns.append(ReplacePattern(re.compile(query[:-1]), lines, 0))

    lines: List[str]
    lines, query, dedent = [], None, 0
    for i, line in enumerate(f_lines):
        line = line.rstrip()

        if line.startswith("#"):
            continue

        if len(line) == 0 or line[0].isspace():
            if not lines:
                stripped = line.lstrip()
                dedent = len(line) - len(stripped)
                lines.append(stripped)
            else:
                s1, s2 = line.lstrip(), line[dedent:]
                lines.append(s2 if len(s2) > len(s1) else s1)
        else:
            if not line.endswith(":"):
                raise Exception(f'Cannot parse line {i+1} of pattern file "{fname}"')

            if query:
                add_pattern(query, lines)
            query = line
            lines = []

    if query:
        add_pattern(query, lines)

    return patterns


def main(args: Optional[List[str]] = None) -> None:
    import sys

    # Ensure that the current directory is on the path
    if "" not in sys.path and "." not in sys.path:
        sys.path.insert(0, "")

    opt = parse_options(sys.argv[1:] if args is None else args)

    patterns: List[ReplacePattern]
    if opt.pattern_file:
        if not opt.quiet:
            print('Using pattern file "%s" ..' % opt.pattern_file)
        patterns = load_pattern_file(opt.pattern_file)
        if not opt.quiet:
            print("  - loaded %i patterns.\n" % len(patterns))
    else:
        patterns = []

    for i in opt.imports:
        sys.path.insert(0, i)

    for i, mod in enumerate(opt.modules):
        if not opt.quiet:
            if i > 0:
                print("\n")
            print('Module "%s" ..' % mod)
            print("  - importing ..")
        mod_imported = importlib.import_module(mod)

        if opt.output_file:
            file = Path(opt.output_file)
        else:
            file_s = getattr(mod_imported, "__file__", None)
            if file_s is None:
                raise Exception(
                    'the module lacks a "__file__" attribute, hence '
                    "stubgen cannot infer where to place the generated "
                    "stub. You must specify the -o parameter to provide "
                    "the name of an output file."
                )
            file = Path(str(file_s))

            ext_loader = importlib.machinery.ExtensionFileLoader
            if isinstance(mod_imported.__loader__, ext_loader):
                # Splitting on "." (module nesting qualifier) handles the case
                # of invoking stubgen on a module that's not in the current
                # working directory - in that case, we still only want the Python
                # module name as the stub file name, not the whole source tree
                # hierarchy.
                modname = mod_imported.__name__.split(".")[-1]
                file = file.with_name(modname)
            file = file.with_suffix(".pyi")

            if opt.output_dir:
                file = Path(opt.output_dir, file.name)

            if opt.recursive and create_subdirectory_for_module(mod_imported) \
                and file.name != '__init__.pyi':
                    file = file.with_suffix('') / "__init__.pyi"

        file.parents[0].mkdir(parents=True, exist_ok=True)

        sg = StubGen(
            module=mod_imported,
            quiet=opt.quiet,
            recursive=opt.recursive,
            include_docstrings=opt.include_docstrings,
            include_private=opt.include_private,
            max_expr_length=0 if opt.exclude_values else 50,
            patterns=patterns,
            output_file=file
        )

        if not opt.quiet:
            print("  - analyzing ..")

        sg.put(mod_imported)

        if patterns:
            total_matches = 0
            for p in patterns:
                if p.matches != 0:
                    total_matches += p.matches
                    continue
                rule_str = str(p.query)
                if "re.compile" in rule_str:
                    rule_str = rule_str.replace("re.compile(", "")[:-1]
                if not opt.quiet:
                    print(f"  - warning: rule {rule_str} did not match any elements.")
            if not opt.quiet:
                print(f"  - applied {total_matches} patterns.")

        file = file.resolve()

        if not opt.quiet:
            print(f'  - writing stub "{file}" ..')

        with open(file, "w", encoding='utf-8') as f:
            f.write(sg.get())

    for marker_file in opt.marker_file:
        marker_file = Path(marker_file).resolve()
        if not opt.quiet:
            print(f'  - writing marker file "{marker_file}" ..')
        marker_file.touch()


if __name__ == "__main__":
    main()



"""
IOWarp Distributed Unit Test Package

Runs distributed bdev unit tests with different PoolQuery strategies:
- DirectHash: Uses loop iterator as hash for distributed allocation
- Range: Uses range query across multiple nodes
- Broadcast: Uses broadcast query to all nodes

Requires wrp_runtime to be running.
"""
from jarvis_cd.core.pkg import Application
from jarvis_cd.shell import Exec, LocalExecInfo
from jarvis_cd.shell.process import Which
import os


class WrpDistributed(Application):
    """
    IOWarp Distributed Unit Test

    Executes distributed bdev tests that verify proper operation of
    PoolQuery strategies (DirectHash, Range, Broadcast) across multiple nodes.

    Parameters:
    - num_nodes: Number of nodes in the distributed system
    - test_case: Which test case to run (direct, range, broadcast, all)
    - output_dir: Directory for test results
    - verbose: Enable verbose test output

    Assumes chimaera_distributed_bdev_tests is installed and available in PATH.
    Requires wrp_runtime to be running.
    """

    def _init(self):
        """Initialize test package variables"""
        self.output_file = None

    def _configure_menu(self):
        """Define configuration options for distributed test"""
        return [
            {
                'name': 'num_nodes',
                'msg': 'Number of distributed nodes',
                'type': int,
                'default': 4
            },
            {
                'name': 'test_case',
                'msg': 'Test case to run (direct, range, broadcast, all)',
                'type': str,
                'choices': ['direct', 'range', 'broadcast', 'all'],
                'default': 'all'
            },
            {
                'name': 'output_dir',
                'msg': 'Output directory for test results',
                'type': str,
                'default': '/tmp/wrp_distributed'
            },
            {
                'name': 'verbose',
                'msg': 'Enable verbose test output',
                'type': bool,
                'default': True
            }
        ]

    def _configure(self, **kwargs):
        """Configure the distributed test"""
        # Create output directory
        os.makedirs(self.config['output_dir'], exist_ok=True)
        self.output_file = os.path.join(self.config['output_dir'], 'distributed_test_results.txt')

        # Set test environment variables
        self.setenv('WRP_DISTRIBUTED_OUTPUT_DIR', self.config['output_dir'])

        self.log("IOWarp distributed test configured")
        self.log(f"  Num nodes: {self.config['num_nodes']}")
        self.log(f"  Test case: {self.config['test_case']}")
        self.log(f"  Output directory: {self.config['output_dir']}")

    def start(self):
        """Run the distributed test"""
        # Verify test executable is available
        Which('chimaera_distributed_bdev_tests', LocalExecInfo(env=self.mod_env)).run()

        self.log(f"Starting distributed bdev test: {self.config['test_case']}")

        # Determine which test cases to run
        test_cases = []
        if self.config['test_case'] == 'all':
            test_cases = ['direct', 'range', 'broadcast']
        else:
            test_cases = [self.config['test_case']]

        # Run each test case
        for test_case in test_cases:
            self.log(f"Running {test_case} test case...")

            # Build test command
            cmd_parts = [
                'chimaera_distributed_bdev_tests',
                f'--num-nodes {self.config["num_nodes"]}',
                f'--test-case {test_case}'
            ]

            if self.config['verbose']:
                cmd_parts.append('--verbose')

            cmd = ' '.join(cmd_parts)

            # Create test-specific output file
            test_output_file = os.path.join(
                self.config['output_dir'],
                f'distributed_test_{test_case}.txt'
            )

            # Redirect output to file and console
            cmd_with_redirect = f'{cmd} 2>&1 | tee {test_output_file}'

            # Execute test
            self.log(f"Executing: {cmd}")
            Exec(cmd_with_redirect, LocalExecInfo(env=self.mod_env)).run()

            self.log(f"Test case '{test_case}' completed - results saved to {test_output_file}")

        self.log(f"All distributed tests completed")

    def stop(self):
        """Stop method - tests complete automatically"""
        pass

    def clean(self):
        """Clean test output"""
        self.log("Cleaning distributed test data")

        # Remove output files
        if self.output_file and os.path.exists(self.output_file):
            os.remove(self.output_file)

        # Remove test-specific output files
        for test_case in ['direct', 'range', 'broadcast']:
            test_output_file = os.path.join(
                self.config['output_dir'],
                f'distributed_test_{test_case}.txt'
            )
            if os.path.exists(test_output_file):
                os.remove(test_output_file)

        # Remove output directory if empty
        try:
            os.rmdir(self.config['output_dir'])
        except OSError:
            pass  # Directory not empty or doesn't exist

        self.log("Cleanup completed")



import wx
import pytest
from hdf_compass.compass_viewer.viewer import CompassApp
from hdf_compass.compass_viewer.frame import BaseFrame, NodeFrame
from hdf_compass.compass_viewer.events import CompassOpenEvent
from hdf_compass import compass_model

class MockStore(compass_model.Store):
    @staticmethod
    def plugin_name():
        return "Mock"
    
    @staticmethod
    def plugin_description():
        return "Mock store for testing"
    
    def __init__(self):
        self._valid = True
    
    def __contains__(self, key):
        return key == "/"
    
    @property
    def url(self):
        return "mock://test"
    
    @property
    def display_name(self):
        return "Mock Store"
    
    @property
    def root(self):
        return None
    
    @property
    def valid(self):
        return self._valid
    
    def gethandlers(self, key=None):
        if key is None:
            return [MockNode]
        if key == "/":
            return [MockNode]
        raise KeyError(key)

class MockNode(compass_model.Node):
    class_kind = "Mock Node"
    
    icons = {16: "hdf_compass/compass_viewer/icons/folder_16.png",
             64: "hdf_compass/compass_viewer/icons/folder_64.png"}
    
    @staticmethod
    def can_handle(store, key):
        return True
    
    def __init__(self, store, key="/"):
        self._store = store
        self._key = key
    
    @property
    def key(self):
        return self._key
    
    @property
    def store(self):
        return self._store
    
    @property
    def display_name(self):
        return "Mock Node"
    
    @property
    def description(self):
        return "Mock node for testing"

def test_wxpython_version():
    """Test that we're using wxPython3"""
    assert wx.version().startswith('4')

def test_compass_app_creation():
    """Test that CompassApp can be created"""
    app = CompassApp(False)
    assert isinstance(app, wx.App)

def test_base_frame_creation():
    """Test that BaseFrame can be created"""
    app = CompassApp(False)
    frame = BaseFrame()
    assert isinstance(frame, wx.Frame)

def test_node_frame_creation():
    """Test that NodeFrame can be created"""
    app = CompassApp(False)
    store = MockStore()
    node = MockNode(store)
    frame = NodeFrame(node)
    assert isinstance(frame, BaseFrame)

def test_compass_open_event():
    """Test that CompassOpenEvent can be created"""
    store = MockStore()
    node = MockNode(store)
    event = CompassOpenEvent(node)
    assert isinstance(event, wx.PyCommandEvent)
    assert event.GetEventType() == CompassOpenEvent.typeId 


##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

""" A setuptools based setup module.

See:
https://packaging.python.org/en/latest/distributing.html
https://github.com/pypa/sampleproject

Run `python setup.py --help-commands` for available options
"""

from __future__ import absolute_import, division, print_function  # unicode_literals

import os
import sys
# To use a consistent encoding
from codecs import open

# Always prefer setuptools over distutils
from setuptools import setup, find_packages

# ---------------------------------------------------------------------------
#                             Some helper stuff
# ---------------------------------------------------------------------------

here = os.path.abspath(os.path.dirname(__file__))


def txt_read(*paths):
    """ Build a file path from *paths* and return the textual contents """
    with open(os.path.join(here, *paths), encoding='utf-8') as f:
        return f.read()

# ---------------------------------------------------------------------------
#                      Populate dictionary with settings
# ---------------------------------------------------------------------------

# Create a dict with the basic information that is passed to setup after keys are added.
setup_args = dict()

setup_args['name'] = 'hdf_compass'
# The adopted versioning scheme follow PEP40
setup_args['version'] = '0.7.0b1'
setup_args['url'] = 'https://github.com/HDFGroup/hdf-compass/'
setup_args['license'] = 'BSD-like license'
setup_args['author'] = 'HDFGroup'
setup_args['author_email'] = 'help@hdfgroup.org'

#
# descriptive stuff
#

description = 'An experimental viewer program for HDF5 and related formats.'
setup_args['description'] = description
setup_args['long_description'] = txt_read('README.rst')

setup_args['classifiers'] = \
    [  # https://pypi.python.org/pypi?%3Aaction=list_classifiers
        'Development Status :: 4 - Beta',
        'Intended Audience :: Science/Research',
        'Natural Language :: English',
        'License :: OSI Approved :: BSD License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
        # 'Programming Language :: Python :: 3',
        # 'Programming Language :: Python :: 3.4',
        'Topic :: Scientific/Engineering :: Information Analysis',
        'Topic :: Office/Business :: Office Suites',
        'Topic :: Utilities'
    ]
setup_args['keywords'] = "data hdf bag ascii grid opendap"

#
# code stuff
#

# requirements
setup_args['setup_requires'] =\
    [
        "setuptools",
        "wheel",
    ]
setup_args['install_requires'] =\
    [
        "numpy>=1.19.0",
        "matplotlib>=3.8.0",
        "h5py>=2.10.0",
        "wxPython>=4.0.0",
        "requests"
    ]
setup_args['extras_require'] =\
    {
        "GeoNodes": ["cartopy[plotting]", ],  # required for visualization of GeoArray and GeoSurface nodes
        "BAG": ["hydroffice.bag>=0.2.10", ],  # required by BAG plugin
        "OpenDAP": ["pydap<3.2", ],  # required by OpenDAP plugin, there is an issue
                                     # with pydap 3.2: https://github.com/pydap/pydap/issues/66
        "ADIOS": ["adios>=1.9.1b19", ],  # required by ADIOS plugin
    }
# hdf_compass namespace, packages and other files
setup_args['namespace_packages'] = ['hdf_compass']
setup_args['packages'] = find_packages(exclude=["*.tests", "*.tests.*", "tests.*", "tests", "*.test*",
                                                ])
setup_args['package_data'] =\
    {
        '': ['icons/*.png', 'icons/*.ico', 'icons/*.icns', 'icons/*.txt'],
    }
setup_args['data_files'] = []
setup_args['entry_points'] =\
    {
        'gui_scripts': ['HDFCompass = hdf_compass.compass_viewer.viewer:run'],
    }

# ---------------------------------------------------------------------------
#                            Do the actual setup now
# ---------------------------------------------------------------------------

setup(**setup_args)



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################
"""
Handles list and icon view for Container display.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import wx
from wx.lib.pubsub import pub

import logging
log = logging.getLogger(__name__)

from hdf_compass import compass_model
from ..events import CompassOpenEvent
from ..events import ContainerSelectionEvent

ID_CONTEXT_MENU_OPEN = wx.ID_ANY
ID_CONTEXT_MENU_OPENWINDOW = wx.ID_ANY
ID_CONTEXT_MENU_COPY = wx.ID_ANY


class ContainerList(wx.ListCtrl):
    """ List control for displaying container contents """

    def __init__(self, parent, node, **kwds):
        """ Create a new container list control """
        default_style = wx.LC_REPORT | wx.LC_SINGLE_SEL | wx.BORDER_SUNKEN
        if 'style' in kwds:
            # Use the passed style, preserving non-conflicting flags
            passed_style = kwds.pop('style')
            # Keep BORDER and SINGLE_SEL but replace the view mode
            style = passed_style | wx.LC_SINGLE_SEL | wx.BORDER_SUNKEN
        else:
            style = default_style
        super(ContainerList, self).__init__(parent, style=style, **kwds)

        self.node = node
        self.parent = parent

        # Create columns only for report view
        if style & wx.LC_REPORT:
            self.InsertColumn(0, "Name")
            self.InsertColumn(1, "Type")
            self.InsertColumn(2, "Size")

            # Set column widths
            self.SetColumnWidth(0, 200)
            self.SetColumnWidth(1, 100)
            self.SetColumnWidth(2, 100)

        # Set size hints
        self.SetMinSize((400, 300))

        # Populate list
        self.populate()

        # Bind events
        self.Bind(wx.EVT_LIST_ITEM_ACTIVATED, self.on_item_activated)
        self.Bind(wx.EVT_LIST_ITEM_RIGHT_CLICK, self.on_right_click)
        self.Bind(wx.EVT_MENU, self.on_copy, id=wx.ID_COPY)
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_item_selected)

    def populate(self):
        """ Populate the list with container contents """
        self.DeleteAllItems()
        for i, child in enumerate(self.node):
            self.InsertItem(i, child.display_name)
            self.SetItem(i, 1, child.__class__.__name__)
            self.SetItem(i, 2, str(child.size))

    def on_item_activated(self, evt):
        """ Handle item activation """
        index = evt.GetIndex()
        try:
            # Check if store is still valid (not closed during shutdown)
            if not hasattr(self.node, 'store') or not self.node.store.valid:
                log.debug("Store is no longer valid, ignoring item activation")
                evt.Skip()
                return
                
            if index < 0 or index >= len(self.node):
                log.warning(f"Item activation index {index} out of bounds")
                evt.Skip()
                return
            child = self.node[index]
            # Get the current position of the parent frame
            pos = self.parent.GetPosition()
            # Create and post the open event with position
            open_event = CompassOpenEvent(child, pos=pos)
            wx.PostEvent(wx.GetApp(), open_event)
        except Exception as e:
            log.exception(f"Error in item activation: {e}")
        evt.Skip()

    def on_item_selected(self, evt):
        """ Handle item selection """
        index = evt.GetIndex()
        try:
            # Check if store is still valid (not closed during shutdown)
            if not hasattr(self.node, 'store') or not self.node.store.valid:
                log.debug("Store is no longer valid, ignoring item selection")
                evt.Skip()
                return
                
            if index < 0 or index >= len(self.node):
                log.warning(f"Item selection index {index} out of bounds")
                evt.Skip()
                return
            child = self.node[index]
            # Create and post the selection event
            selection_event = ContainerSelectionEvent(child)
            wx.PostEvent(self.parent, selection_event)
        except Exception as e:
            log.exception(f"Error in item selection: {e}")
        evt.Skip()

    def on_right_click(self, evt):
        """ Handle right click """
        menu = wx.Menu()
        menu.Append(wx.ID_COPY, "Copy")
        self.PopupMenu(menu)
        menu.Destroy()

    def on_copy(self, evt):
        """ Handle copy menu item """
        index = self.GetFirstSelected()
        if index != -1:
            child = self.node[index]
            pub.sendMessage('compass.copy', node=child)

    def copy_to_clipboard(self):
        """ Copy selected item to clipboard """
        index = self.GetFirstSelected()
        if index != -1:
            child = self.node[index]
            pub.sendMessage('compass.copy', node=child)

    @property
    def selection(self):
        """ Return the currently selected node, or None """
        try:
            # Check if store is still valid (not closed during shutdown)
            if not hasattr(self.node, 'store') or not self.node.store.valid:
                log.debug("Store is no longer valid, returning None for selection")
                return None
                
            index = self.GetFirstSelected()
            if index != -1:
                return self.node[index]
            return None
        except Exception as e:
            log.debug(f"Error getting selection (likely due to shutdown): {e}")
            return None


class ContainerIconList(ContainerList):
    """
    Icon view of nodes in a Container.
    """

    def __init__(self, parent, node):
        """ New icon list view
        """
        style = wx.LC_ICON | wx.LC_AUTOARRANGE | wx.BORDER_SUNKEN
        super(ContainerIconList, self).__init__(parent, node, style=style)
        
        self.node = node

        # Set minimum size for icon view
        self.SetMinSize((600, 400))

        # Set up image list for icons
        app = wx.GetApp()
        if not hasattr(app, 'imagelists') or 64 not in app.imagelists:
            # Ensure imagelists are initialized
            app.init_imagelists()
        self.il = app.imagelists[64]
        self.SetImageList(self.il, wx.IMAGE_LIST_NORMAL)

        # Clear any existing items
        self.DeleteAllItems()

        # Populate with icons
        for item in range(len(self.node)):
            try:
                subnode = self.node[item]
                if hasattr(self, 'il') and self.il is not None:
                    image_index = self.il.get_index(type(subnode))
                    self.InsertItem(item, subnode.display_name, image_index)
                else:
                    # Fallback without icon if image list not available
                    self.InsertItem(item, subnode.display_name)
            except:
                log.exception("Error adding icon item")

        # Force initial layout
        self.Layout()
        self.Refresh()

    def populate(self):
        """ Override populate to handle icon view """
        # Ensure image list is available
        if not hasattr(self, 'il') or self.il is None:
            app = wx.GetApp()
            if not hasattr(app, 'imagelists') or 64 not in app.imagelists:
                app.init_imagelists()
            self.il = app.imagelists[64]
            self.SetImageList(self.il, wx.IMAGE_LIST_NORMAL)
        
        self.DeleteAllItems()
        for item in range(len(self.node)):
            try:
                subnode = self.node[item]
                if hasattr(self, 'il') and self.il is not None:
                    image_index = self.il.get_index(type(subnode))
                    self.InsertItem(item, subnode.display_name, image_index)
                else:
                    # Fallback without icon if image list not available
                    self.InsertItem(item, subnode.display_name)
            except:
                log.exception("Error adding icon item")
        self.Layout()
        self.Refresh()


class ContainerReportList(ContainerList):
    """
    List view of the container's contents.

    Uses a wxPython virtual list, allowing millions of items in a container
    without any slowdowns.
    """

    def __init__(self, parent, node):
        """ Create a new list view.

        parent: wxPython parent object
        node:   Container instance to be displayed
        """

        style = wx.LC_REPORT | wx.LC_VIRTUAL | wx.LC_SINGLE_SEL | wx.BORDER_SUNKEN
        super(ContainerReportList, self).__init__(parent, node, style=style)

        self.node = node

        # Clear and recreate columns for report view
        self.ClearAll()
        self.InsertColumn(0, "Name")
        self.InsertColumn(1, "Kind")
        self.SetColumnWidth(0, 300)
        self.SetColumnWidth(1, 200)

        # Set minimum size
        self.SetMinSize((500, 400))

        # Set up image list
        app = wx.GetApp()
        if not hasattr(app, 'imagelists') or 16 not in app.imagelists:
            # Ensure imagelists are initialized
            app.init_imagelists()
        self.il = app.imagelists[16]
        self.SetImageList(self.il, wx.IMAGE_LIST_SMALL)

        # Set up virtual list
        self.SetItemCount(len(node))

        # Note: EVT_LIST_ITEM_ACTIVATED is already bound by parent ContainerList class
        # Only bind the selection event which might need different handling for virtual list
        self.Bind(wx.EVT_LIST_ITEM_SELECTED, self.on_item_selected)

        # Force initial layout
        self.Layout()
        self.Refresh()

    def OnGetItemText(self, item, col):
        """ Callback method to support virtual list ctrl """
        try:
            # Check if store is still valid (not closed during shutdown)
            if not hasattr(self.node, 'store') or not self.node.store.valid:
                return ""
                
            # Check bounds first
            if item < 0 or item >= len(self.node):
                log.warning(f"Item index {item} out of bounds (length: {len(self.node)})")
                return ""
                
            subnode = self.node[item]
            if col == 0:
                return subnode.display_name
            elif col == 1:
                return type(subnode).class_kind
        except Exception as e:
            log.debug(f"Error getting item text for item {item}, col {col} (likely due to shutdown): {e}")
        return ""

    def OnGetItemImage(self, item):
        """ Callback method to support virtual list ctrl """
        try:
            # Check if store is still valid (not closed during shutdown)
            if not hasattr(self.node, 'store') or not self.node.store.valid:
                return -1
                
            # Check bounds first
            if item < 0 or item >= len(self.node):
                log.warning(f"Item index {item} out of bounds (length: {len(self.node)})")
                return -1
                
            subnode = self.node[item]
            if hasattr(self, 'il') and self.il is not None:
                return self.il.get_index(type(subnode))
            return -1
        except Exception as e:
            log.debug(f"Error getting item image for item {item} (likely due to shutdown): {e}")
            return -1

    def populate(self):
        """ Override populate to use virtual list """
        try:
            # Check if store is still valid (not closed during shutdown)
            if not hasattr(self.node, 'store') or not self.node.store.valid:
                log.debug("Store is no longer valid, setting item count to 0")
                self.SetItemCount(0)
                return
                
            node_length = len(self.node)
            log.debug(f"Setting virtual list item count to {node_length}")
            self.SetItemCount(node_length)
            self.Refresh()
        except Exception as e:
            log.exception(f"Error in populate: {e}")
            self.SetItemCount(0)



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Implements a viewer for compass_model.Container instances.

This frame is a simple browser view with back/forward/up controls.

Currently list and icon views are supported.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import wx
import os

import logging
log = logging.getLogger(__name__)

from hdf_compass import compass_model
from ..frame import NodeFrame
from ..events import ID_COMPASS_OPEN
from ..events import EVT_CONTAINER_SELECTION
from .list import ContainerReportList, ContainerIconList

ID_GO_MENU_BACK = wx.NewId()
ID_GO_MENU_NEXT = wx.NewId()
ID_GO_MENU_UP = wx.NewId()
ID_GO_MENU_TOP = wx.NewId()

ID_VIEW_MENU_LIST = wx.NewId()
ID_VIEW_MENU_ICON = wx.NewId()


class ContainerFrame(NodeFrame):
    """
    A frame to display a Container class, and browse its contents.
    """

    def __init__(self, node, pos=None):
        """ Create a new frame.

        node:   Container instance to display.
        pos:    Screen position at which to display the window.
        """
        super(ContainerFrame, self).__init__(node, size=(800, 400), title=node.display_title, pos=pos)

        # Create view and set it using the view property
        self.view = ContainerReportList(self, node)

        view_menu = wx.Menu()
        view_menu.Append(ID_VIEW_MENU_LIST, "List view")
        view_menu.Append(ID_VIEW_MENU_ICON, "Icon view")
        self.add_menu(view_menu, "View")
        self.view_menu = view_menu

        go_menu = wx.Menu()
        go_menu.Append(ID_GO_MENU_BACK, "Back")
        go_menu.Append(ID_GO_MENU_NEXT, "Next")
        go_menu.Append(ID_GO_MENU_UP, "Up")
        go_menu.Append(ID_GO_MENU_TOP, "Top")
        self.add_menu(go_menu, "Go")
        self.go_menu = go_menu

        self.acc_tbl = wx.AcceleratorTable([
            (wx.ACCEL_CMD, wx.WXK_LEFT, ID_GO_MENU_BACK),
            (wx.ACCEL_CMD, wx.WXK_RIGHT, ID_GO_MENU_NEXT),
            (wx.ACCEL_CMD, wx.WXK_UP, ID_GO_MENU_UP),
            (wx.ACCEL_CMD | wx.ACCEL_SHIFT, wx.WXK_UP, ID_GO_MENU_TOP),
        ])
        self.SetAcceleratorTable(self.acc_tbl)

        self.Bind(wx.EVT_MENU, self.on_compass_open, id=ID_COMPASS_OPEN)
        self.Bind(EVT_CONTAINER_SELECTION, lambda evt: self.update_info())

        self.Bind(wx.EVT_MENU, lambda evt: self.go_back(), id=ID_GO_MENU_BACK)
        self.Bind(wx.EVT_MENU, lambda evt: self.go_next(), id=ID_GO_MENU_NEXT)
        self.Bind(wx.EVT_MENU, lambda evt: self.go_up(), id=ID_GO_MENU_UP)
        self.Bind(wx.EVT_MENU, lambda evt: self.go_top(), id=ID_GO_MENU_TOP)
        self.Bind(wx.EVT_MENU, lambda evt: self.list_view(), id=ID_VIEW_MENU_LIST)
        self.Bind(wx.EVT_MENU, lambda evt: self.icon_view(), id=ID_VIEW_MENU_ICON)

        # Only create toolbar if it doesn't exist
        existing_toolbar = self.GetToolBar()
        if existing_toolbar:
            self.toolbar = existing_toolbar
        else:
            self.toolbar = self.CreateToolBar(wx.TB_HORIZONTAL | wx.NO_BORDER | wx.TB_FLAT | wx.TB_TEXT)

        tsize = (24, 24)
        back_bmp = wx.Bitmap(os.path.join(self.icon_folder, "go_back_24.png"), wx.BITMAP_TYPE_ANY)
        next_bmp = wx.Bitmap(os.path.join(self.icon_folder, "go_next_24.png"), wx.BITMAP_TYPE_ANY)
        up_bmp = wx.Bitmap(os.path.join(self.icon_folder, "go_up_24.png"), wx.BITMAP_TYPE_ANY)
        top_bmp = wx.Bitmap(os.path.join(self.icon_folder, "go_top_24.png"), wx.BITMAP_TYPE_ANY)
        icon_bmp = wx.Bitmap(os.path.join(self.icon_folder, "view_icon_24.png"), wx.BITMAP_TYPE_ANY)
        list_bmp = wx.Bitmap(os.path.join(self.icon_folder, "view_list_24.png"), wx.BITMAP_TYPE_ANY)

        self.toolbar.SetToolBitmapSize(tsize)
        self.toolbar.AddTool(ID_GO_MENU_BACK, "Back", back_bmp, "Go back to previous location")
        self.toolbar.AddTool(ID_GO_MENU_NEXT, "Next", next_bmp, "Go forward to next location")
        self.toolbar.AddSeparator()
        self.toolbar.AddTool(ID_GO_MENU_UP, "Up", up_bmp, "Go up one level")
        self.toolbar.AddTool(ID_GO_MENU_TOP, "Top", top_bmp, "Go to root level")
        self.toolbar.AddStretchableSpace()
        self.toolbar.AddTool(ID_VIEW_MENU_LIST, "List View", list_bmp, "Switch to list view")
        self.toolbar.AddTool(ID_VIEW_MENU_ICON, "Icon View", icon_bmp, "Switch to icon view")

        self.toolbar.Realize()

        self.history = [node]
        self.history_ptr = 0
        self.update_view()
        
        # Ensure the frame is shown and raised
        self.Show()
        self.Raise()

    def list_view(self):
        """ Switch to list view """
        if not isinstance(self.view, ContainerReportList):
            self.view = ContainerReportList(self, self.history[self.history_ptr])

    def icon_view(self):
        """ Switch to icon view """
        if not isinstance(self.view, ContainerIconList):
            self.view = ContainerIconList(self, self.history[self.history_ptr])

    # --- Begin history support functions -------------------------------------

    @property
    def node(self):
        return self.history[self.history_ptr]

    def go_back(self):
        """ Go to the previously displayed item """
        if self.history_ptr > 0:
            self.history_ptr -= 1
        self.update_view()

    def go_next(self):
        """ Go to the next displayed item """
        if self.history_ptr < (len(self.history) - 1):
            self.history_ptr += 1
        self.update_view()

    def go_up(self):
        """ Go to the enclosing container """
        node = self.history[self.history_ptr]
        parent = node.store.get_parent(node.key)
        if parent.key != node.key:  # at the root item
            self.go(parent)

    def go_top(self):
        """ Go to the root node """
        node = self.history[self.history_ptr]
        parent = node.store.root
        if parent.key != node.key:  # at the root item
            self.go(parent)

    def go(self, newnode):
        """ Go to a particular node.

        Adds the node to the history.
        """
        del self.history[self.history_ptr + 1:]
        self.history.append(newnode)
        self.history_ptr += 1
        self.update_view()

    # --- End history support functions ---------------------------------------

    @property
    def node(self):
        """ Get the current node from history """
        if hasattr(self, 'history') and self.history and hasattr(self, 'history_ptr'):
            return self.history[self.history_ptr]
        else:
            # Fallback to parent implementation during initialization
            return super(ContainerFrame, self).node

    def update_view(self):
        """ Refresh the entire contents of the frame according to current history node. """
        self.SetTitle(self.node.display_title)
        self.view = type(self.view)(self, self.node)
        self.update_info()

        can_go_back = self.history_ptr > 0
        can_go_next = self.history_ptr < (len(self.history) - 1)
        can_go_up = self.node.store.get_parent(self.node.key) is not None
        can_go_top = self.node.key != self.node.store.root.key
        self.go_menu.Enable(ID_GO_MENU_BACK, can_go_back)
        self.go_menu.Enable(ID_GO_MENU_NEXT, can_go_next)
        self.go_menu.Enable(ID_GO_MENU_UP, can_go_up)
        self.go_menu.Enable(ID_GO_MENU_TOP, can_go_top)
        self.toolbar.EnableTool(ID_GO_MENU_BACK, can_go_back)
        self.toolbar.EnableTool(ID_GO_MENU_NEXT, can_go_next)
        self.toolbar.EnableTool(ID_GO_MENU_UP, can_go_up)
        self.toolbar.EnableTool(ID_GO_MENU_TOP, can_go_top)

        icon_view_allowed = len(self.node) <= 1000
        self.view_menu.Enable(ID_VIEW_MENU_ICON, icon_view_allowed)
        self.toolbar.EnableTool(ID_VIEW_MENU_ICON, icon_view_allowed)

    def update_info(self):
        """ Refresh the left-hand side information panel, according to the
        current selection in the view.
        """
        node = self.view.selection
        if node is None:
            node = self.node  # Nothing selected; show this node's info
        self.info.display(node)

    def on_compass_open(self, evt):
        """ Handle compass open event.
        
        If the node is a container, navigate to it in this frame.
        Otherwise, let the event propagate to create a new window.
        """
        try:
            # Check if this is a CompassOpenEvent with a node attribute
            if hasattr(evt, 'node'):
                new_node = evt.node
                log.debug("Got request to open node: %s" % new_node.key)
                if isinstance(new_node, compass_model.Container):
                    self.go(new_node)
                else:
                    evt.Skip()  # Let the event propagate to create a new window
            else:
                # This might be a menu event when no specific node is selected
                # Check if there's a current selection in the view
                if hasattr(self.view, 'selection') and self.view.selection is not None:
                    selected_node = self.view.selection
                    log.debug("Opening selected node: %s" % selected_node.key)
                    if isinstance(selected_node, compass_model.Container):
                        self.go(selected_node)
                    else:
                        # Create an open event for the selected node to open in new window
                        log.debug(f"Creating new window for non-container node: {selected_node.key} (type: {type(selected_node)})")
                        from ..events import CompassOpenEvent
                        open_event = CompassOpenEvent(selected_node, pos=self.GetPosition())
                        log.debug(f"Posting CompassOpenEvent for node: {selected_node.key}")
                        wx.PostEvent(wx.GetApp(), open_event)
                else:
                    log.debug("No node selected for open action")
                    evt.Skip()
        except Exception as e:
            log.exception(f"Error in on_open: {e}")
            evt.Skip()



from jarvis_cd.core.pkg import Service
from jarvis_cd.util import SizeType
from jarvis_cd.shell.process import Rm, Mkdir
from jarvis_cd.shell import PsshExecInfo, Exec, LocalExecInfo
import yaml
import os

class WrpCte(Service):
    """
    Content Transfer Engine (CTE) configuration service for IoWarp.
    
    This service configures the CTE core module by generating YAML configuration
    files and setting appropriate environment variables. It does not run as a
    persistent service but rather prepares the environment for CTE usage.
    """
    
    def _init(self):
        """
        Initialize the WrpCte service.
        
        This method is called during service initialization.
        """
        self.config_file_path = None
        self.devices_from_resource_graph = []

    def _configure_menu(self):
        """
        Configure the service menu.
        
        Returns:
            List[Dict]: Configuration menu options for CTE.
        """
        return [
            {
                'name': 'devices',
                'msg': 'List of storage devices as tuples (path, capacity, score)',
                'type': list,
                'default': [],
                'help': 'Example: [("/mnt/nvme", "1TB", 0.9), ("ram::cache", "8GB", 1.0)]. Use ram:: prefix for RAM storage. Supports SizeType format: k/K, m/M, g/G, t/T'
            },
            {
                'name': 'dpe_type',
                'msg': 'Data Placement Engine type',
                'type': str,
                'choices': ['random', 'round_robin', 'max_bw'],
                'default': 'max_bw'
            },
            {
                'name': 'neighborhood',
                'msg': 'Number of targets (nodes CTE can buffer to)',
                'type': int,
                'default': 4
            },
            {
                'name': 'default_target_timeout_ms',
                'msg': 'Default target timeout in milliseconds',
                'type': int,
                'default': 30000
            },
            {
                'name': 'poll_period_ms',
                'msg': 'Period at which targets should be rescanned for statistics (capacity, bandwidth, etc.) in milliseconds',
                'type': int,
                'default': 5000
            }
        ]

    def _configure(self, **kwargs):
        """
        Configure the CTE service with provided keyword arguments.
        
        This method generates the CTE YAML configuration file and sets
        the WRP_RUNTIME_CONF environment variable.
        
        Args:
            **kwargs: Configuration arguments from _configure_menu.
        """
        self.log("Configuring Content Transfer Engine (CTE)...")
        
        # Handle devices configuration
        devices = self.config.get('devices', [])
        
        if not devices:
            self.log("No devices specified, attempting to use resource graph...")
            devices = self._get_devices_from_resource_graph()
            if not devices:
                self.log("Warning: No devices available from resource graph, using defaults")
                devices = self._get_default_devices()
        else:
            # Validate and convert device tuples
            devices = self._validate_and_convert_devices(devices)
        
        
        # Build CTE configuration
        cte_config = self._build_cte_config(devices)
        
        # Save configuration to shared directory
        self.config_file_path = os.path.join(self.shared_dir, 'cte_config.yaml')
        
        try:
            with open(self.config_file_path, 'w') as f:
                yaml.dump(cte_config, f, default_flow_style=False, indent=2)
            self.log(f"CTE configuration written to: {self.config_file_path}")
        except Exception as e:
            self.log(f"Error writing CTE configuration: {e}")
            raise
        
        # Set environment variable
        self.setenv('WRP_RUNTIME_CONF', self.config_file_path)
        self.log(f"Environment variable WRP_RUNTIME_CONF set to: {self.config_file_path}")

        # Create parent directories for each device
        self.log("Creating parent directories for storage devices...")
        for path, capacity, score in devices:
            parent_dir = os.path.dirname(path)
            self.log(f"Creating directory: {parent_dir}")
            try:
                Mkdir(parent_dir, PsshExecInfo(hostfile=self.hostfile)).run()
                self.log(f"Created directory: {parent_dir}")
            except Exception as e:
                self.log(f"Error creating directory {parent_dir}: {e}")

        self.log("CTE configuration completed successfully")

    def _get_devices_from_resource_graph(self):
        """
        Extract device information from the resource graph.

        Returns:
            List[Tuple[str, str, float]]: List of device tuples (path, capacity, score).
        """
        try:
            from jarvis_cd.core.resource_graph import ResourceGraphManager

            # Initialize ResourceGraphManager (gets Jarvis singleton internally)
            rg_manager = ResourceGraphManager()

            if not rg_manager.resource_graph.get_all_nodes():
                self.log("Warning: Resource graph is empty. Run 'jarvis rg build' first.")
                return []

            # Get common storage - returns dict mapping mount points to device lists
            common_storage = rg_manager.resource_graph.get_common_storage()
            if not common_storage:
                self.log("Warning: No common storage found in resource graph")
                return []

            devices = []
            # Iterate over common storage mount points
            for mount_point, device_list in common_storage.items():
                # Process each device at this mount point
                for device in device_list:
                    # Use mount point as base path
                    base_path = device.get('mount', '/tmp/cte_storage')
                    # Append hermes_data.bin for bdev file creation
                    path = os.path.join(base_path, 'hermes_data.bin')

                    # Get available space and multiply by 0.5 for safety margin
                    available_space = device.get('avail', '100GB')
                    capacity = self._adjust_capacity(available_space, 0.5)

                    # Calculate score based on device type
                    device_type = device.get('dev_type', 'unknown').lower()
                    score = self._calculate_device_score(device_type, device)

                    devices.append((path, capacity, score))
                    self.log(f"Found storage device: {path} (available: {available_space}, using: {capacity}, score: {score}, type: {device_type})")

            return devices

        except ImportError as e:
            self.log(f"Warning: ResourceGraphManager not available: {e}")
            return []
        except Exception as e:
            self.log(f"Warning: Error accessing resource graph: {e}")
            return []

    def _adjust_capacity(self, capacity_str, factor):
        """
        Adjust capacity by a factor (e.g., multiply by 0.5 for safety margin).

        Args:
            capacity_str (str): Capacity string (e.g., "100GB", "1TB")
            factor (float): Multiplication factor

        Returns:
            str: Adjusted capacity string in same format
        """
        import re

        # Parse capacity string
        match = re.match(r'([\d.]+)\s*([KMGT]?B?)', capacity_str.upper().strip())
        if not match:
            # If parsing fails, return original
            return capacity_str

        value = float(match.group(1))
        suffix = match.group(2)

        # Apply factor
        adjusted_value = value * factor

        # Format back to string with 2 decimal places if needed
        if adjusted_value == int(adjusted_value):
            return f"{int(adjusted_value)}{suffix}"
        else:
            return f"{adjusted_value:.2f}{suffix}"

    def _calculate_device_score(self, storage_type, storage_info):
        """
        Calculate a performance score for a storage device.

        Args:
            storage_type (str): Type of storage (nvme, ssd, hdd, ram, etc.)
            storage_info (dict): Additional storage information from resource graph.

        Returns:
            float: Score between 0.0 and 1.0.
        """
        # Base scores by storage type
        type_scores = {
            'ram': 1.0,
            'ramdisk': 1.0,
            'tmpfs': 1.0,
            'nvme': 0.9,
            'ssd': 0.7,
            'hdd': 0.4,
            'network': 0.3,
            'unknown': 0.5
        }

        base_score = type_scores.get(storage_type, 0.5)

        # Adjust based on performance benchmarks from resource graph
        # Resource graph provides '4k_randwrite_bw' and '1m_seqwrite_bw' if benchmarked
        seq_bw = storage_info.get('1m_seqwrite_bw', 'unknown')
        rand_bw = storage_info.get('4k_randwrite_bw', 'unknown')

        # Boost score for high sequential bandwidth
        if seq_bw != 'unknown' and isinstance(seq_bw, str):
            try:
                # Parse bandwidth strings like "500MB/s" or "1.5GB/s"
                if 'GB/s' in seq_bw:
                    bw_value = float(seq_bw.replace('GB/s', '').strip())
                    base_score = min(1.0, base_score + 0.2)  # High performance boost
                elif 'MB/s' in seq_bw:
                    bw_value = float(seq_bw.replace('MB/s', '').strip())
                    if bw_value > 500:  # Over 500 MB/s is good
                        base_score = min(1.0, base_score + 0.1)
            except (ValueError, AttributeError):
                pass

        # Boost score for high random write performance
        if rand_bw != 'unknown' and isinstance(rand_bw, str):
            try:
                if 'MB/s' in rand_bw:
                    bw_value = float(rand_bw.replace('MB/s', '').strip())
                    if bw_value > 50:  # Over 50 MB/s for 4K random is excellent
                        base_score = min(1.0, base_score + 0.1)
            except (ValueError, AttributeError):
                pass

        return round(base_score, 2)

    def _get_default_devices(self):
        """
        Get default device configuration when no resource graph is available.
        
        Returns:
            List[Tuple[str, str, float]]: Default device configuration.
        """
        return [
            ('/tmp/cte_storage/cte_target.bin', '100GB', 0.6)
        ]

    def _validate_and_convert_devices(self, devices):
        """
        Validate and convert device specifications.
        
        Args:
            devices (List): Device specifications to validate.
            
        Returns:
            List[Tuple[str, str, float]]: Validated device tuples.
        """
        validated_devices = []
        
        for i, device in enumerate(devices):
            try:
                if isinstance(device, (list, tuple)) and len(device) >= 3:
                    path, capacity, score = device[0], device[1], device[2]
                    
                    # Validate path
                    if not isinstance(path, str) or not path.strip():
                        raise ValueError(f"Invalid path in device {i}: {path}")
                    
                    # Validate capacity using SizeType
                    capacity_str = self._normalize_capacity_with_sizetype(capacity)
                    
                    # Validate score
                    score_float = float(score)
                    if not 0.0 <= score_float <= 1.0:
                        raise ValueError(f"Score must be between 0.0 and 1.0, got: {score_float}")
                    
                    validated_devices.append((path.strip(), capacity_str, score_float))
                else:
                    raise ValueError(f"Device {i} must be a tuple/list with at least 3 elements")
                    
            except Exception as e:
                self.log(f"Warning: Invalid device specification {i}: {device} - {e}")
                continue
        
        if not validated_devices:
            self.log("Warning: No valid devices found, using defaults")
            return self._get_default_devices()
        
        return validated_devices

    def _normalize_capacity_with_sizetype(self, capacity):
        """
        Normalize capacity specification using Jarvis SizeType utility.
        
        Args:
            capacity: Capacity specification (string or number).
            
        Returns:
            str: Normalized capacity string.
        """
        try:
            # Convert capacity to SizeType - it handles the string formatting automatically
            size_type = SizeType(capacity)
            return str(size_type)
                
        except Exception as e:
            raise ValueError(f"Invalid capacity format '{capacity}': {e}")

    def _build_cte_config(self, devices):
        """
        Build the complete CTE configuration dictionary.
        
        Args:
            devices (List[Tuple]): List of device tuples.
            
        Returns:
            dict: Complete CTE configuration.
        """
        # Convert devices to storage configuration format
        storage_config = []
        for path, capacity, score in devices:
            # Determine bdev_type: check if path begins with ram:: prefix
            if path.startswith('ram::'):
                bdev_type = 'ram'
            else:
                bdev_type = 'file'

            storage_config.append({
                'path': path,
                'bdev_type': bdev_type,
                'capacity_limit': capacity,
                'score': score
            })
        
        # Build complete configuration
        config = {
            'targets': {
                'neighborhood': self.config.get('neighborhood', 4),
                'default_target_timeout_ms': self.config.get('default_target_timeout_ms', 30000),
                'poll_period_ms': self.config.get('poll_period_ms', 5000)
            },

            'storage': storage_config,

            'dpe': {
                'dpe_type': self.config.get('dpe_type', 'max_bw')
            },

            'logging': {
                'level': 'info',
                'file_path': '/var/log/cte/cte.log'
            }
        }
        
        return config

    def start(self):
        """
        Start the WrpCte service by launching CTE using launch_cte.

        This method executes the launch_cte utility to initialize the
        Content Transfer Engine with the generated configuration.

        Returns:
            bool: True if launch_cte executed successfully, False otherwise.
        """
        self.log("Starting Content Transfer Engine...")

        try:
            # Execute launch_cte using LocalExec
            exec_info = LocalExecInfo(env=self.mod_env)
            cmd = "launch_cte"

            self.log(f"Executing: {cmd}")
            Exec(cmd, exec_info).run()
            self.log("Content Transfer Engine launched successfully")
            return True

        except Exception as e:
            self.log(f"Error launching CTE: {e}")
            return False

    def stop(self):
        """
        Stop the WrpCte service.
        
        Since this is a configuration-only service, this method just passes.
        
        Returns:
            bool: Always True since this is configuration-only.
        """
        self.log("WrpCte is a configuration service - no persistent process to stop")
        return True

    def kill(self):
        """
        Force stop the WrpCte service.
        
        Since this is a configuration-only service, this method just passes.
        
        Returns:
            bool: Always True since this is configuration-only.
        """
        self.log("WrpCte is a configuration service - no persistent process to kill")
        return True

    def status(self):
        """
        Check the current status of the WrpCte service.
        
        Returns:
            dict: A dictionary containing service status information.
        """
        if self.config_file_path and os.path.exists(self.config_file_path):
            return {
                "running": True,
                "details": "Configuration service active",
                "config_file": self.config_file_path,
                "env_var_set": 'WRP_RUNTIME_CONF' in (self.env or {})
            }
        else:
            return {
                "running": False,
                "details": "Configuration not generated",
                "config_file": None,
                "env_var_set": False
            }

    def clean(self):
        """
        Clean up CTE configuration files and storage devices using Rm with PsshExecInfo.
        """
        self.log("Starting CTE cleanup process...")
        
        # Clean up configuration file
        if self.config_file_path and os.path.exists(self.config_file_path):
            try:
                os.remove(self.config_file_path)
                self.log(f"Removed CTE configuration file: {self.config_file_path}")
            except Exception as e:
                self.log(f"Error removing configuration file: {e}")
        
        # Clean up storage devices using Rm with PsshExecInfo
        try:
            # Get all configured storage devices
            devices = self.config.get('devices', [])

            # If no devices were manually configured, get from resource graph
            if not devices:
                devices.extend(self._get_devices_from_resource_graph())

            # Clean up each device and parent directory using Rm with PsshExecInfo
            if devices:
                self.log(f"Cleaning up {len(devices)} storage devices...")

                for path, capacity, score in devices:
                    try:
                        # Execute removal using Rm with PsshExecInfo across all nodes
                        Rm(path, PsshExecInfo(hostfile=self.hostfile)).run()
                        self.log(f"Successfully cleaned storage device: {path}")

                        # Remove parent directory
                        parent_dir = os.path.dirname(path)
                        self.log(f"Removing directory: {parent_dir}")
                        Rm(parent_dir, PsshExecInfo(hostfile=self.hostfile)).run()
                        self.log(f"Successfully removed parent directory: {parent_dir}")

                    except Exception as e:
                        self.log(f"Error cleaning storage device {path}: {e}")
            else:
                self.log("No storage devices to clean")

        except Exception as e:
            self.log(f"Error during storage device cleanup: {e}")
        
        self.log("CTE configuration cleanup completed")



#!/usr/bin/env python3
"""
Python bindings test and usage examples for WRP CTE Core - MCP Integration

This script serves as a reference implementation for using the WRP CTE Core
Python bindings in Model Context Protocol (MCP) servers.

Key Features:
-----------
1. Runtime Initialization: Demonstrates how to initialize the Chimaera runtime
   and CTE subsystem with proper error handling

2. Context Operations: Examples of context_bundle, context_query, and context_delete
   - context_bundle: Store data in blobs under tags
   - context_query: List and retrieve blobs from tags
   - context_delete: Delete blobs from tags

Usage:
------
    # Run with runtime initialization (default)
    python3 test_bindings.py

    # Run without runtime initialization
    CHIMAERA_WITH_RUNTIME=0 python3 test_bindings.py

Example Usage in Your Code:
---------------------------
    import wrp_cte_core_ext as cte

    # 1. Initialize runtime (if not already done externally)
    cte.chimaera_init(cte.ChimaeraMode.kClient, True)
    cte.initialize_cte(config_path, cte.PoolQuery.Dynamic())
    
    # 2. Bundle data (context_bundle)
    tag = cte.Tag("my_tag")
    tag.PutBlob("my_blob", b"Hello, World!", 0)
    
    # 3. Query blobs in a tag (context_query - list)
    blobs = tag.GetContainedBlobs()
    
    # 4. Retrieve blob data (context_query - get)
    blob_size = tag.GetBlobSize("my_blob")
    data = tag.GetBlob("my_blob", blob_size, 0)
    
    # 5. Delete blob (context_delete)
    # Note: DelBlob may need to be accessed via Client if not bound to Tag wrapper

Environment Variables:
---------------------
    CHIMAERA_WITH_RUNTIME: Set to "0" or "false" to skip runtime initialization
    CHI_SERVER_CONF: Path to Chimaera server configuration file
    CHI_REPO_PATH: Path to ChiMod repository (for finding shared libraries)
    LD_LIBRARY_PATH: Library path for runtime dependencies
"""

import sys
import os
import time
import signal

# Add current directory to path for module import
sys.path.insert(0, os.getcwd())

# Global state tracking for runtime initialization
runtime_initialized = False
client_initialized = False

# Track if we're attempting initialization
_initialization_attempted = False


def should_initialize_runtime():
    """Check if runtime should be initialized

    Reads CHIMAERA_WITH_RUNTIME environment variable:
    - Not set or "1"/"true"/"yes"/"on": Initialize runtime (default: true)
    - "0"/"false"/"no"/"off": Skip initialization (runtime already initialized externally)
    """
    # Check unified flag
    env_val = os.getenv("CHIMAERA_WITH_RUNTIME")
    if env_val is None:
        return True  # Default: initialize runtime

    # Case-insensitive check for false values
    env_val_lower = str(env_val).lower()
    return env_val_lower not in ("0", "false", "no", "off")


def setup_environment_paths():
    """Set up CHI_REPO_PATH and LD_LIBRARY_PATH for ChiMod discovery (following C++ test pattern)
    
    This is critical for the runtime to find ChiMod shared libraries.
    Gets the build directory by finding where the Python module is located.
    """
    try:
        # Try to find the module's location
        import wrp_cte_core_ext as cte
        module_file = cte.__file__ if hasattr(cte, '__file__') else None
        
        if module_file:
            # Get the directory containing the module
            bin_dir = os.path.dirname(os.path.abspath(module_file))
            
            # Set CHI_REPO_PATH and LD_LIBRARY_PATH to point to bin directory
            os.environ["CHI_REPO_PATH"] = bin_dir
            
            # Update LD_LIBRARY_PATH, preserving existing path
            existing_ld_path = os.getenv("LD_LIBRARY_PATH", "")
            if existing_ld_path:
                os.environ["LD_LIBRARY_PATH"] = f"{bin_dir}:{existing_ld_path}"
            else:
                os.environ["LD_LIBRARY_PATH"] = bin_dir
            
            print(f"   Set CHI_REPO_PATH={bin_dir}")
            print(f"   Set LD_LIBRARY_PATH={os.environ['LD_LIBRARY_PATH']}")
            return True
    except Exception as e:
        print(f"   ⚠️  Could not determine module path: {e}")
        # Try to use current working directory as fallback
        cwd = os.getcwd()
        if os.path.exists(cwd):
            os.environ["CHI_REPO_PATH"] = cwd
            existing_ld_path = os.getenv("LD_LIBRARY_PATH", "")
            if existing_ld_path:
                os.environ["LD_LIBRARY_PATH"] = f"{cwd}:{existing_ld_path}"
            else:
                os.environ["LD_LIBRARY_PATH"] = cwd
            print(f"   Set CHI_REPO_PATH={cwd} (fallback)")
            return True
    return False


def generate_test_config():
    """Generate a minimal test configuration for Chimaera runtime

    Example: Configuration File Structure
    ------------------------------------
    This demonstrates how to create a Chimaera configuration file programmatically.
    The configuration is written as YAML and should contain:
    
    - networking: Protocol, hostfile, and port settings
    - workers: Number of worker threads
    - memory: Segment sizes for main, client, and runtime
    - devices: Storage device mount points and capacities
    
    Example config structure:
        {
            'networking': {
                'protocol': 'zmq',
                'hostfile': '/path/to/hostfile',
                'port': 9129
            },
            'workers': {
                'num_workers': 4
            },
            'memory': {
                'main_segment_size': '1G',
                'client_data_segment_size': '512M',
                'runtime_data_segment_size': '512M'
            },
            'devices': [
                {
                    'mount_point': '/path/to/storage',
                    'capacity': '1G'
                }
            ]
        }
    
    Creates a YAML config file with proper networking and storage settings.
    Returns the path to the generated config file.
    """
    import tempfile
    import socket

    def find_available_port(start_port=9129, end_port=9200):
        """Find an available port in the given range"""
        for port in range(start_port, end_port):
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                try:
                    s.bind(('', port))
                    return port
                except OSError:
                    continue
        raise RuntimeError(f"No available ports in range {start_port}-{end_port}")

    try:
        import yaml
    except ImportError:
        print("⚠️  PyYAML not available - cannot generate config")
        return None

    temp_dir = tempfile.gettempdir()

    # Create clean hostfile
    clean_hostfile = os.path.join(temp_dir, "wrp_bindings_test_hostfile")
    with open(clean_hostfile, 'w') as f:
        f.write("localhost\n")

    # Find available port
    port = find_available_port()
    print(f"   Using port: {port}")

    # Create storage directory
    storage_dir = os.path.join(temp_dir, "cte_bindings_test_storage")
    os.makedirs(storage_dir, exist_ok=True)

    # Generate config
    config = {
        'networking': {
            'protocol': 'zmq',
            'hostfile': clean_hostfile,
            'port': port
        },
        'workers': {
            'num_workers': 4
        },
        'memory': {
            'main_segment_size': '1G',
            'client_data_segment_size': '512M',
            'runtime_data_segment_size': '512M'
        },
        'devices': [
            {
                'mount_point': storage_dir,
                'capacity': '1G'
            }
        ]
    }

    # Write config
    config_path = os.path.join(temp_dir, "wrp_bindings_test_conf.yaml")
    with open(config_path, 'w') as f:
        yaml.dump(config, f)

    print(f"   Generated config: {config_path}")

    # Set environment variable
    os.environ['CHI_SERVER_CONF'] = config_path

    return config_path


def initialize_runtime_early(cte):
    """Initialize Chimaera runtime early in the test (called from main() before client code)

    Example: Complete Runtime Initialization Pattern
    ------------------------------------------------
    This demonstrates the complete pattern for initializing the Chimaera runtime
    and CTE subsystem. This is required before any client operations.
    
    Usage Pattern:
        # Step 1: Set up environment paths (for ChiMod discovery)
        os.environ["CHI_REPO_PATH"] = "/path/to/build/bin"
        os.environ["LD_LIBRARY_PATH"] = "/path/to/build/bin"
        
        # Step 2: Generate or load configuration
        config_path = generate_config()  # Creates YAML config with networking, storage, etc.
        os.environ["CHI_SERVER_CONF"] = config_path
        
        # Step 3: Initialize Chimaera (unified init - both runtime and client)
        if not cte.chimaera_init(cte.ChimaeraMode.kClient, True):
            raise RuntimeError("Failed to initialize Chimaera")
        time.sleep(0.5)  # Give Chimaera time to initialize
        
        # Step 5: Initialize CTE subsystem
        pool_query = cte.PoolQuery.Dynamic()
        if not cte.initialize_cte(config_path, pool_query):
            raise RuntimeError("Failed to initialize CTE subsystem")
    
    This follows the C++ test pattern from context-runtime/test/unit/test_chimaera_runtime.cc:
    - Calls CHIMAERA_INIT(ChimaeraMode::kClient, true)
    - Sleeps 500ms after initialization
    - Verifies initialization state

    Returns True if successful, False otherwise.
    """
    global runtime_initialized, client_initialized, _initialization_attempted

    if runtime_initialized:
        print("✅ Runtime already initialized")
        # Still try to initialize client if not done
        if client_initialized:
            return True
    else:
        _initialization_attempted = True

    try:
        # Step 0: Generate test configuration
        print("🔧 Generating test configuration...")
        config_path = generate_test_config()
        if not config_path:
            print("⚠️  Could not generate test configuration")
            print("   Continuing with binding tests only...")
            return False

        # Step 1: Initialize Chimaera (unified init - both runtime and client)
        # Following pattern from test_chimaera_runtime.cc
        if not runtime_initialized or not client_initialized:
            print("🔧 Initializing Chimaera (unified CHIMAERA_INIT)...")
            print("   Note: If runtime isn't configured, this may cause FATAL and process exit")
            sys.stdout.flush()  # Ensure output is flushed before potential abort

            try:
                init_result = cte.chimaera_init(cte.ChimaeraMode.kClient, True)
            except Exception as e:
                print(f"⚠️  Chimaera initialization exception: {e}")
                print("   Continuing with binding tests only...")
                return False

            if not init_result:
                print("⚠️  Chimaera initialization returned False")
                print("   This may indicate runtime configuration issues")
                print("   Continuing with binding tests only...")
                return False

            runtime_initialized = True
            client_initialized = True

            # Give Chimaera time to initialize all components (following C++ pattern: 500ms)
            time.sleep(0.5)

            # Verify initialization succeeded
            print("✅ Chimaera initialized")
            sys.stdout.flush()

            # Verify client initialization (following C++ pattern that checks IPC)
            # In C++ tests they verify: REQUIRE(CHI_IPC != nullptr) and REQUIRE(CHI_IPC->IsInitialized())
            print("✅ Chimaera client initialized")
            sys.stdout.flush()

        # Step 3: Initialize CTE subsystem (CTE-specific, not in base runtime tests)
        print("🔧 Initializing CTE subsystem...")
        sys.stdout.flush()

        try:
            pool_query = cte.PoolQuery.Dynamic()
            cte_result = cte.initialize_cte(config_path, pool_query)
        except Exception as e:
            print(f"⚠️  CTE initialization exception: {e}")
            print("   Continuing with binding tests only...")
            return False

        if not cte_result:
            print("⚠️  CTE initialization returned False")
            print("   This may indicate CTE configuration issues")
            print("   Continuing with binding tests only...")
            return False

        print("✅ CTE subsystem initialized")
        sys.stdout.flush()
        
        # Step 4: Register a storage target (required for PutBlob operations)
        print("🔧 Registering storage target...")
        try:
            client = cte.get_cte_client()
            mctx = cte.MemContext()
            
            # Get storage directory from config (or use a default)
            import yaml
            with open(config_path, 'r') as f:
                config = yaml.safe_load(f)
            storage_dir = config.get('devices', [{}])[0].get('mount_point', '/tmp/cte_test_storage')
            
            # Create target path
            target_path = os.path.join(storage_dir, "test_target")
            os.makedirs(os.path.dirname(target_path), exist_ok=True)
            
            # Register file-based target (1GB size) with explicit PoolId
            # Use a high pool ID (700) to avoid conflicts with system pools
            bdev_id = cte.PoolId(700, 0)
            target_query = cte.PoolQuery.Local()
            result = client.RegisterTarget(mctx, target_path, cte.BdevType.kFile, 
                                           1024 * 1024 * 1024, target_query, bdev_id)
            
            if result == 0:
                print("✅ Storage target registered successfully")
            else:
                print(f"⚠️  Storage target registration returned {result} (may be expected if already registered)")
        except Exception as e:
            print(f"⚠️  Could not register storage target: {e}")
            print("   PutBlob operations may fail without registered targets")
        
        return True

    except SystemExit as e:
        # Handle FATAL errors from C++ that cause SystemExit
        print(f"⚠️  Runtime initialization FATAL error (process would exit): {e}")
        print("   This usually means Chimaera runtime isn't properly configured")
        print("   Continuing with binding tests only...")
        sys.stdout.flush()
        return False
    except Exception as e:
        print(f"⚠️  Runtime initialization error: {e}")
        print("   Continuing with binding tests only...")
        import traceback
        traceback.print_exc()
        sys.stdout.flush()
        return False


def test_context_bundle_operation(cte):
    """Test context bundle operation (equivalent to context_bundle function)
    
    Example: Bundling data into a blob under a tag
    -----------------------------------------------
    This demonstrates how to use Tag.PutBlob() to bundle data, which is the
    underlying operation for context_bundle(tag_name, blob_name, data).
    
    Usage Pattern:
        tag = cte.Tag(tag_name)
        tag.PutBlob(blob_name, data_bytes, offset=0)
    
    Parameters:
        tag_name (str): Name of the tag to store the blob under
        blob_name (str): Name of the blob to create
        data_bytes (bytes): Binary data to store in the blob
        offset (int): Optional byte offset (default: 0)
    
    Returns:
        None (raises exception on failure)
    """
    global runtime_initialized, client_initialized
    
    if not runtime_initialized or not client_initialized:
        print("⚠️  Skipping context_bundle test (runtime not initialized)")
        return True  # Not a failure, just skipped
    
    try:
        print("🔧 Testing context_bundle operation (Tag.PutBlob)...")
        
        # Create a test tag
        test_tag_name = "test_context_bundle_tag"
        test_blob_name = "test_bundle_blob"
        test_data = b"Hello from context_bundle test!"
        
        try:
            # Use Tag wrapper to bundle data (equivalent to context_bundle)
            tag = cte.Tag(test_tag_name)
            tag.PutBlob(test_blob_name, test_data, 0)
            
            print(f"   ✅ context_bundle operation succeeded")
            print(f"   Bundled data: tag='{test_tag_name}', blob='{test_blob_name}', size={len(test_data)} bytes")
            return True
        except Exception as e:
            print(f"   ⚠️  context_bundle operation failed: {e}")
            # Don't fail the test if it fails - may be expected without proper setup
            return True
        
    except Exception as e:
        print(f"⚠️  context_bundle test error (may be expected): {e}")
        return True


def test_context_query_operations(cte):
    """Test context query operations (equivalent to context_query function)
    
    Example: Querying blobs in a tag
    ---------------------------------
    This demonstrates two query patterns:
    
    1. List all blobs in a tag (equivalent to context_query(tag_name)):
       Usage:
           tag = cte.Tag(tag_name)
           blob_list = tag.GetContainedBlobs()  # Returns list[str]
    
    2. Retrieve specific blob data (equivalent to context_query(tag_name, blob_name)):
       Usage:
           tag = cte.Tag(tag_name)
           blob_size = tag.GetBlobSize(blob_name)  # Get size first
           data = tag.GetBlob(blob_name, blob_size, offset=0)  # Retrieve data
           # Returns str/bytes containing blob data
    """
    global runtime_initialized, client_initialized
    
    if not runtime_initialized or not client_initialized:
        print("⚠️  Skipping context_query tests (runtime not initialized)")
        return True  # Not a failure, just skipped
    
    try:
        test_tag_name = "test_context_query_tag"
        
        # Test 1: List blobs (equivalent to context_query(tag_name))
        print("🔧 Testing context_query (list blobs) - Tag.GetContainedBlobs()...")
        try:
            tag = cte.Tag(test_tag_name)
            
            # Try to bundle some test blobs first
            try:
                tag.PutBlob("query_blob1", b"Data 1", 0)
                tag.PutBlob("query_blob2", b"Data 2", 0)
            except Exception:
                pass  # May fail if runtime isn't fully set up
            
            # Query all blobs in tag
            blob_list = tag.GetContainedBlobs()
            
            assert isinstance(blob_list, list), "GetContainedBlobs should return a list"
            if blob_list:
                assert all(isinstance(blob, str) for blob in blob_list), \
                    "GetContainedBlobs should return list of strings"
                print(f"   ✅ context_query (list) returned {len(blob_list)} blobs: {blob_list}")
            else:
                print(f"   ⚠️  context_query (list) returned empty list (may be expected)")
            
        except Exception as e:
            print(f"   ⚠️  context_query (list) failed: {e}")
        
        # Test 2: Get specific blob (equivalent to context_query(tag_name, blob_name))
        print("🔧 Testing context_query (get blob) - Tag.GetBlob()...")
        try:
            tag = cte.Tag(test_tag_name)
            test_blob_name = "query_get_blob"
            test_data = b"This is test data for retrieval!"
            
            # Bundle data first
            try:
                tag.PutBlob(test_blob_name, test_data, 0)
            except Exception:
                print("   ⚠️  Could not bundle test data (may be expected)")
                return True
            
            # Get blob size first
            blob_size = tag.GetBlobSize(test_blob_name)
            if blob_size > 0:
                # Retrieve blob data
                retrieved_data = tag.GetBlob(test_blob_name, blob_size, 0)
                
                assert isinstance(retrieved_data, (str, bytes)), \
                    "GetBlob should return str or bytes"
                
                # Convert to bytes for comparison
                if isinstance(retrieved_data, str):
                    retrieved_bytes = retrieved_data.encode('latin-1')
                else:
                    retrieved_bytes = retrieved_data
                
                if retrieved_bytes == test_data:
                    print(f"   ✅ context_query (get blob) retrieved correct data ({len(retrieved_bytes)} bytes)")
                else:
                    print(f"   ⚠️  context_query (get blob) data mismatch")
            else:
                print(f"   ⚠️  Blob size is 0, cannot retrieve")
                
        except Exception as e:
            print(f"   ⚠️  context_query (get blob) failed: {e}")
        
        print("✅ context_query tests completed")
        return True
        
    except Exception as e:
        print(f"⚠️  context_query test error (may be expected): {e}")
        return True


def test_context_delete_operation(cte):
    """Test context delete operation (equivalent to context_delete function)
    
    Example: Deleting a blob from a tag
    ------------------------------------
    Tests DelBlob operation to delete a blob from a tag.
    This is the underlying operation for context_delete(tag_name, blob_name).
    Uses Client.DelBlob() since DelBlob is not bound to Tag wrapper.
    
    Usage Pattern:
        # Get tag and its TagId
        tag = cte.Tag(tag_name)
        tag_id = tag.GetTagId()
        
        # Delete blob using Client
        client = cte.get_cte_client()
        mctx = cte.MemContext()
        result = client.DelBlob(mctx, tag_id, blob_name)
        # Returns bool: True if successful, False otherwise
    """
    global runtime_initialized, client_initialized
    
    if not runtime_initialized or not client_initialized:
        print("⚠️  Skipping context_delete test (runtime not initialized)")
        return True  # Not a failure, just skipped
    
    try:
        print("🔧 Testing context_delete operation (Client.DelBlob)...")
        
        # Check if DelBlob is available on Client
        client = cte.get_cte_client()
        client_type = cte.Client
        
        # DelBlob is on Client, not Tag wrapper
        if hasattr(client_type, 'DelBlob'):
            print("   ✅ Client.DelBlob method found")
            # Test deletion
            try:
                test_tag_name = "test_context_delete_tag"
                test_blob_name = "delete_test_blob"
                
                # Create tag and bundle data first
                tag = cte.Tag(test_tag_name)
                tag_id = tag.GetTagId()
                
                # Bundle data first
                try:
                    tag.PutBlob(test_blob_name, b"Data to delete", 0)
                    print(f"   ✅ Bundled test data for deletion")
                except Exception as e:
                    print(f"   ⚠️  Could not bundle test data: {e}")
                    return True
                
                # Delete blob using Client.DelBlob
                mctx = cte.MemContext()
                delete_result = client.DelBlob(mctx, tag_id, test_blob_name)
                
                if delete_result:
                    print(f"   ✅ context_delete operation succeeded")
                else:
                    print(f"   ⚠️  context_delete returned False (blob may not exist)")
                
            except Exception as e:
                print(f"   ⚠️  context_delete operation failed: {e}")
                import traceback
                traceback.print_exc()
        else:
            print("   ⚠️  DelBlob method not available on Client")
            print("   Note: context_delete may not be fully implemented yet")
        
        print("✅ context_delete test completed")
        return True
        
    except Exception as e:
        print(f"⚠️  context_delete test error (may be expected): {e}")
        import traceback
        traceback.print_exc()
        return True


def main():
    """Run all context operation tests"""
    print("=" * 70)
    print("🧪 Python Bindings Test Suite - MCP Integration Examples")
    print("=" * 70)
    print()
    
    # STEP 0: Runtime initialization (if enabled) - MUST BE FIRST before any client code
    # Following pattern from context-runtime/test/unit/test_chimaera_runtime.cc
    # Runtime initialization happens at the very beginning if enabled
    runtime_ok = False
    if should_initialize_runtime():
        print("📋 Initializing Runtime (CHIMAERA_WITH_RUNTIME enabled)...")
        print("   Note: Runtime initialization happens FIRST before any client code")

        # Import module first (needed for runtime init)
        try:
            import wrp_cte_core_ext as cte
        except ImportError as e:
            print(f"❌ Cannot import module for runtime init: {e}")
            return 1

        # Set up environment paths for ChiMod discovery (before runtime init)
        # Following pattern from context-transfer-engine/test/unit/test_query.cc:114-136
        print("🔧 Setting up environment paths for ChiMod discovery...")
        setup_environment_paths()

        # Initialize runtime NOW (before any client code)
        # Following pattern from context-runtime/test/unit/test_chimaera_runtime.cc:58-84
        runtime_ok = initialize_runtime_early(cte)
        print()
    else:
        cte_flag = os.getenv("CHIMAERA_WITH_RUNTIME")
        if cte_flag:
            print(f"📋 Skipping Runtime Initialization (CHIMAERA_WITH_RUNTIME={cte_flag})")
        else:
            print("📋 Skipping Runtime Initialization")
        print("   Runtime should already be initialized externally")
        print()
    
    # Import module
    try:
        import wrp_cte_core_ext as cte
    except ImportError as e:
        print(f"❌ Cannot import module: {e}")
        return 1
    
    # Test 1: Context bundle operation (context_bundle equivalent)
    if runtime_ok:
        print("📋 Test 1: Context Bundle Operation (context_bundle)...")
        test_context_bundle_operation(cte)
        print()
        
        # Test 2: Context query operations (context_query equivalent)
        print("📋 Test 2: Context Query Operations (context_query)...")
        test_context_query_operations(cte)
        print()
        
        # Test 3: Context delete operation (context_delete equivalent)
        print("📋 Test 3: Context Delete Operation (context_delete)...")
        test_context_delete_operation(cte)
        print()
    else:
        print("⚠️  Skipping context operation tests (runtime not initialized)")
        print()
    
    # Summary
    print("=" * 70)
    print("📊 Test Summary")
    print("=" * 70)
    print("✅ All context operation tests completed")
    if runtime_ok:
        print("✅ Runtime tests executed")
    else:
        print("⚠️  Runtime tests skipped (runtime not initialized)")
    print()
    print("🎉 Python bindings test suite passed!")
    print("=" * 70)
    
    return 0


if __name__ == "__main__":
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        print("\n\n⚠️  Interrupted by user")
        sys.exit(1)
    except Exception as e:
        print(f"\n\n❌ Error: {e}")
        import traceback
        traceback.print_exc()
        sys.exit(1)



##############################################################################
# Copyright by The HDF Group.                                                #
# All rights reserved.                                                       #
#                                                                            #
# This file is part of the HDF Compass Viewer. The full HDF Compass          #
# copyright notice, including terms governing use, modification, and         #
# terms governing use, modification, and redistribution, is contained in     #
# the file COPYING, which can be found at the root of the source code        #
# distribution tree.  If you do not have access to this file, you may        #
# request a copy from help@hdfgroup.org.                                     #
##############################################################################

"""
Defines a small number of custom events, which are useful for the GUI.
"""
from __future__ import absolute_import, division, print_function, unicode_literals

import wx
from wx.lib.newevent import NewEvent

import logging
log = logging.getLogger(__name__)

ID_COMPASS_OPEN = wx.ID_ANY

# Create new event type
_CompassOpenEvent, EVT_COMPASS_OPEN = NewEvent()

class CompassOpenEvent(_CompassOpenEvent):
    """ Event sent when a node should be opened """
    def __init__(self, node, pos=None):
        super(CompassOpenEvent, self).__init__()
        self.node = node
        self.pos = pos

    def get_node(self):
        """ Get the node to open """
        return self.node

    """
    Event posted when a request has been made to "open" a object in
    the container.  The source should always be ID_COMPASS_OPEN.

    The type of event is EVT_MENU, because wxPython doesn't like it if we
    use anything else.  When binding handlers, make sure to explicitly
    specify the source ID (or check it in the callback) to avoid catching
    these events by mistake.
    """

# Hints that the selected item in the view may have changed.
# Interested parties should inspect the view to figure out the new selection.
_ContainerSelectionEvent, EVT_CONTAINER_SELECTION = NewEvent()

class ContainerSelectionEvent(_ContainerSelectionEvent):
    """ Event sent when container selection changes """
    def __init__(self, node=None):
        super(ContainerSelectionEvent, self).__init__()
        self.node = node
